////////////////////////////////////////////////////////////////////////
/*
     FILE: AVHRRData.java
  PURPOSE: Interface for data from radiometers.
   AUTHOR: Peter Hollemans
     DATE: 2007/02/23
  CHANGES: n/a

  Copyright (C) 2007, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.instrument;

// Import 
// ------
import java.io.*;
import java.nio.*;
import terrenus.telemetry.*;
import terrenus.instrument.RadiometerCalibrator.Record;
import terrenus.instrument.RadiometerCalibrator.CalibrationType;

/**
 * The <code>AVHRRData</code> class holds data from an AVHRR/3 instrument
 * on the NOAA KLM series spacecrafts. 
 *
 * @see AVHRR
 * @see terrenus.orbit.NOAAKLMSpacecraft
 */
public class AVHRRData
  implements RadiometerData {

  // Variables
  // ---------

  /** The frame number for this data. */
  private int frame;

  /** The decoder to use for frame data. */
  private HRPTDecoder decoder;

  /** The calibrator to use for calibration. */
  private AVHRRCalibrator calibrator;

  /** The calibration record for this data. */
  private Record record;

  /** The location generator. */
  private AVHRRLocator locator;

  /** The temporary count data array. */
  private int[] tempCountData = new int[AVHRR.SAMPLES];

  /** The sample values array. */
  private static int[] sampleValues;

  ////////////////////////////////////////////////////////////

  /**
   * Creates a new AVHRR data object for the specified telemetry frame.
   *
   * @param frame the frame number for the data.
   * @param decoder the HRPT decoder to use for data.
   * @param calibrator the calibrator to use for calibration.
   * @param record the calibration record for this data or null
   * if one cannot be created.
   * @param locator the location generator.
   */
  public AVHRRData (
    int frame,
    HRPTDecoder decoder,
    AVHRRCalibrator calibrator,
    Record record,
    AVHRRLocator locator
  ) {

    this.frame = frame;
    this.decoder = decoder;
    this.calibrator = calibrator;
    this.record = record;
    this.locator = locator;

  } // AVHRRData constructor

  ////////////////////////////////////////////////////////////

  public Instrument getInstrument() { return (AVHRR.getInstance()); }

  ////////////////////////////////////////////////////////////

  public int[] getCountData (
    int channel, 
    int[] countData
  ) throws IOException {

    // Create count data array
    // -----------------------
    if (countData == null) countData = new int[AVHRR.SAMPLES];

    // Determine if channel is valid
    // -----------------------------
    boolean is3a = decoder.isChannel3a (frame);
    if (channel == 3 && is3a)
      throw new IllegalArgumentException ("No channel 3b data available");
    else if (channel == 6 && !is3a)
      throw new IllegalArgumentException ("No channel 3a data available");

    // Get count data for channel
    // --------------------------
    ShortBuffer earthData = decoder.getStream().getBlock (frame, 
      HRPTStream.EARTH_BLOCK).asShortBuffer();
    int bufOffset = (channel == 6 ? 2 : (channel-1));
    for (int i = 0; i < AVHRR.SAMPLES; i++)
      countData[i] = earthData.get (i*5 + bufOffset);

    return (countData);

  } // getCountData

  ////////////////////////////////////////////////////////////

  public double[] getCalibratedData (
    int channel,
    CalibrationType calType,
    double[] calData
  ) throws IOException {

    // Check calibration record
    // ------------------------
    if (record == null)
      throw new IOException ("No calibration record available");

    // Create calibrated data array
    // ----------------------------
    if (calData == null) calData = new double[AVHRR.SAMPLES];

    // Get count data and calibrate
    // ----------------------------
    getCountData (channel, tempCountData);
    calibrator.calibrate (tempCountData, calData, channel, calType, record);

    return (calData);

  } // getCalibratedData

  ////////////////////////////////////////////////////////////

  public void getLocationData (
    double[] latitude,
    double[] longitude,
    double[] satZenith,
    double[] solZenith,
    double[] relAzimuth
  ) {

    // Create sample values array
    // --------------------------
    if (sampleValues == null) {
      sampleValues = new int[AVHRR.SAMPLES];
      for (int i = 0; i < AVHRR.SAMPLES; i++) sampleValues[i] = i;
    } // if

    // Compute location and angle data
    // -------------------------------
    locator.getLocations (frame, sampleValues, latitude, longitude,
      satZenith, solZenith, relAzimuth);

  } // getLocationData

  ////////////////////////////////////////////////////////////

} // AVHRRData class

////////////////////////////////////////////////////////////////////////
