////////////////////////////////////////////////////////////////////////
/*
     FILE: AbstractRadiometerCalibrator.java
  PURPOSE: Interface for calibrating radiometer data.
   AUTHOR: Peter Hollemans
     DATE: 2007/03/03
  CHANGES: n/a

  Copyright (C) 2007, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.instrument;

// Imports
// -------
import java.io.*;
import java.util.*;

/**
 * The <code>AbstractRadiometerCalibrator</code> class implements
 * a scheme for organizing radiometer calibration records.  A
 * call to the {@link #getCalibration(int)} method performs
 * calibration on a group of consecutive scan lines and caches
 * the record for later use.  If no calibration can be performed
 * for a given group of lines, the nearest valid calibration
 * record (up to some limited number of line away) is used
 * instead.
 */
public abstract class AbstractRadiometerCalibrator 
  implements RadiometerCalibrator {

  // Variables
  // ---------

  /** The total number of scan lines. */
  private int lines;

  /** The cache calibration records indexed by the start line. */
  private Map<Integer,Record> recordCacheMap;

  /** The set of failed records that cannot be computed. */
  private Set<Integer> failedRecordSet;

  /** The number of scan lines used to compute a calibration record. */
  private int recordLines;

  /** The maximum number of records away to copy a record. */
  private int recordCopyLimit;

  ////////////////////////////////////////////////////////////

  /**
   * Creates a new calibrator.
   *
   * @param lines the number of scan lines to be calibrated.
   * @param recordLines the number of scan lines to use to compute
   * each calibration record.
   * @param recordCopyLimit the maximum number of records away to
   * copy a record when calibration cannot be computed for a
   * record.
   */
  public AbstractRadiometerCalibrator (
    int lines,
    int recordLines,
    int recordCopyLimit
  ) {

    // Initialize
    // ----------
    this.lines = lines;
    this.recordLines = recordLines;
    this.recordCopyLimit = recordCopyLimit;

    // Create record cache
    // -------------------
    recordCacheMap = new HashMap<Integer,Record>();
    failedRecordSet = new HashSet<Integer>();

  } // AbstractRadiometerCalibrator constructor

  ////////////////////////////////////////////////////////////

  /** 
   * Gets a calibration record from the cache.  No attempt is
   * made to search for a nearby record.
   *
   * @param startLine the starting line for the calibration record.
   * 
   * @return the cached record or null if no record can be computed.
   */
  private Record getCachedCalibration (
    int startLine
  ) throws IOException {

    // Check in record cache
    // ---------------------
    Record record = recordCacheMap.get (startLine);

    // Create new record if possible
    // -----------------------------
    if (record == null && !failedRecordSet.contains (startLine)) {
      int endLine = Math.min (startLine+recordLines-1, lines-1);
      record = getCalibration (startLine, endLine);
      if (record != null)
        recordCacheMap.put (startLine, record);
      else
        failedRecordSet.add (startLine);
    } // if

    return (record);

  } // getCachedCalibration

  ////////////////////////////////////////////////////////////

  public Record getCalibration (
    int line
  ) throws IOException {

    // Check in record cache
    // ---------------------
    int startLine = (line / recordLines) * recordLines;
    Record record = getCachedCalibration (startLine);

    // Search surrounding records
    // --------------------------
    if (record == null) {
      for (int diff = 1; diff < recordCopyLimit; diff++) {

        // Look ahead
        // ----------
        int newStartLine = startLine + diff*recordLines;
        if (newStartLine <= lines-1) {
          record = getCachedCalibration (newStartLine);
          if (record != null) break;
        } // if

        // Look behind
        // -----------
        newStartLine = startLine - diff*recordLines;
        if (newStartLine >= 0) {
          record = getCachedCalibration (newStartLine);
          if (record != null) break;
        } // if

      } // for
    } // if

    return (record);

  } // getCalibration

  ////////////////////////////////////////////////////////////

} // AbstractRadiometerCalibrator interface

////////////////////////////////////////////////////////////////////////
