////////////////////////////////////////////////////////////////////////
/*
     File: CWNCDataWriter.java
  PURPOSE: Writes radiometer data to CW-style netCDF format.
   AUTHOR: Peter Hollemans
     DATE: 2007/08/14
  CHANGES: 2010/12/01, PFH, added log inhibit calls
           2011/02/05, PFH
           - added writeMissingLineData method to write missing data
           - added check for Double.NaN values when writing data

  Copyright (C) 2007-2011, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.instrument;

// Imports
// -------
import java.io.*;
import java.util.*;
import ucar.nc2.*;
import ucar.ma2.*;
import terrenus.util.*;
import terrenus.telemetry.*;
import java.util.logging.*;

/**
 * The <code>CWNCDataWriter</code> class writes radiometer data
 * to netCDF version 3 files.  All available radiometer channels
 * are written in units specified by the {@link #writeData}
 * method.  The metadata written conforms to the CoastWatch HDF
 * version 3.4 specificaion.  Latitude, longitude, satellite
 * zenith, solar zenith, and relative azimuth data are also
 * computed and written in degrees.  Latitude and longitude
 * variables are written as netCDF 32-bit floats, and all other
 * angle data and radiometer channels as netCDF 16-bit signed
 * integers scaled by 0.01.
 */
public class CWNCDataWriter 
  extends AbstractRadiometerDataWriter {

  // Constants
  // ---------

  /** The number of milliseconds in one day. */
  private static final int MSEC_PER_DAY = 86400000;

  /** The set of variable names to write as float data. */
  private static final HashSet<String> FLOAT_VARS = 
    new HashSet<String> (Arrays.asList (new String[] {
      "latitude", 
      "longitude"
    }));

  // Variables
  // ---------

  /** The output file name. */
  private String output;

  /** The netCDF writeable file. */
  private NetcdfFileWriteable ncFile;
  
  /** The variable name array. */
  private String[] varNameArray;

  /** The array of flags that indicate whether data should be scaled. */
  private boolean[] isScaled;

  /** The array of temporary floating point values for writing. */
  private float[] floatData;

  /** The array of temporary short values for writing. */
  private short[] shortData;

  /** The data value count for writing each data line. */
  private int[] count;

  /** The data value start for writing each data line. */
  private int[] start;

  ////////////////////////////////////////////////////////////

  /** Creates a new writer. */
  public CWNCDataWriter (
    String output
  ) {

    this.output = output;

  } // CWNCDataWriter constuctor

  ////////////////////////////////////////////////////////////

  protected void createVariables (
    List<String> varNames,
    List<String> varUnits,
    int[] dims
  ) throws IOException {

    // Create dimensions
    // -----------------
    Dimension[] ncDims = new Dimension[] {
      ncFile.addDimension ("row", dims[0]),
      ncFile.addDimension ("column", dims[1])
    };

    // Loop over each variable
    // -----------------------
    int variables = varNames.size();
    varNameArray = new String[variables];
    isScaled = new boolean[variables];
    for (int i = 0; i < variables; i++) {

      // Set variable properties
      // -----------------------
      String varName = varNames.get (i);
      varNameArray[i] = varName;
      String varUnit = varUnits.get (i);
      isScaled[i] = !FLOAT_VARS.contains (varName);

      // Create integer variable
      // -----------------------
      if (isScaled[i]) {
        ncFile.addVariable (varName, DataType.SHORT, ncDims);
        ncFile.addVariableAttribute (varName, "long_name", varName);
        ncFile.addVariableAttribute (varName, "units", varUnit);
        ncFile.addVariableAttribute (varName, "fraction_digits", 2);
        ncFile.addVariableAttribute (varName, "missing_value", 
          Short.MIN_VALUE);
        ncFile.addVariableAttribute (varName, "_FillValue", Short.MIN_VALUE);
        ncFile.addVariableAttribute (varName, "scale_factor", 0.01);
        ncFile.addVariableAttribute (varName, "scale_factor_err", 0);
        ncFile.addVariableAttribute (varName, "add_offset", 0);
        ncFile.addVariableAttribute (varName, "add_offset_err", 0);
        ncFile.addVariableAttribute (varName, "calibrated_nt", 0);
      } // if

      // Create float variable
      // ---------------------
      else {
        ncFile.addVariable (varName, DataType.FLOAT, ncDims);
        ncFile.addVariableAttribute (varName, "long_name", varName);
        ncFile.addVariableAttribute (varName, "units", varUnit);
        ncFile.addVariableAttribute (varName, "fraction_digits", 6);
        ncFile.addVariableAttribute (varName, "missing_value", Float.NaN);
        ncFile.addVariableAttribute (varName, "_FillValue", Float.NaN);
      } // else

    } // for

    // Put file into data mode
    // -----------------------
    ncFile.create();

  } // createVariables

  ////////////////////////////////////////////////////////////

  protected void writeMissingLineData (
    int varIndex,
    int line
  ) throws IOException {
  
    // Prepare missing output values
    // -----------------------------
    Object data;
    if (isScaled[varIndex]) {
      for (int i = 0; i < shortData.length; i++)
        shortData[i] = Short.MIN_VALUE;
      data = shortData;
    } // if
    else {
      for (int i = 0; i < floatData.length; i++)
        floatData[i] = Float.NaN;
      data = floatData;
    } // else
    
    // Write data
    // ----------
    start[0] = line;
    try {
      ncFile.write (varNameArray[varIndex], start, 
        Array.factory (data.getClass().getComponentType(), count, data));
    } // try
    catch (InvalidRangeException e) {
      throw new IOException (e.toString());
    } // catch

  } // writeMissingLineData

  ////////////////////////////////////////////////////////////

  protected void writeLineData (
    int varIndex,
    int line,
    double[] lineData
  ) throws IOException {

    // Convert data to output values
    // -----------------------------
    Object data;
    if (isScaled[varIndex]) {
      for (int i = 0; i < lineData.length; i++) {
        if (Double.isNaN (lineData[i]))
          shortData[i] = Short.MIN_VALUE;
        else
          shortData[i] = (short) Math.round (lineData[i] * 100);
      } // for
      data = shortData;
    } // if
    else {
      for (int i = 0; i < lineData.length; i++) {
        if (Double.isNaN (lineData[i]))
          floatData[i] = Float.NaN;
        else
          floatData[i] = (float) lineData[i];
      } // for
      data = floatData;
    } // else
    
    // Write data
    // ----------
    start[0] = line;
    try {
      ncFile.write (varNameArray[varIndex], start, 
        Array.factory (data.getClass().getComponentType(), count, data));
    } // try
    catch (InvalidRangeException e) {
      throw new IOException (e.toString());
    } // catch

  } // writeLineData

  ////////////////////////////////////////////////////////////

  protected void openOutput (
    TelemetryDecoder decoder,
    Radiometer radiometer,
    int startFrame,
    int endFrame
  ) throws IOException {

    // Turn off NetCDF logging
    // -----------------------
    LogManager manager = java.util.logging.LogManager.getLogManager();
    Logger logger = manager.getLogger ("");
    logger.setLevel (Level.OFF);

    // Create file
    // -----------
    ncFile = NetcdfFileWriteable.createNew (output, false);

    // Set global attributes
    // ---------------------
    ncFile.addGlobalAttribute ("satellite", decoder.getSpacecraft().getName());
    ncFile.addGlobalAttribute ("sensor", radiometer.getName());
    ncFile.addGlobalAttribute ("origin", origin);
    ncFile.addGlobalAttribute ("cwhdf_version", "3.4");
    ncFile.addGlobalAttribute ("pass_type", "unknown");
    long msec = decoder.getTime (startFrame);
    ncFile.addGlobalAttribute ("pass_date", (int) (msec/MSEC_PER_DAY));
    ncFile.addGlobalAttribute ("start_time", (msec%MSEC_PER_DAY)/1000.0);
    ncFile.addGlobalAttribute ("projection_type", "swath");
    int lines = endFrame - startFrame + 1;
    ncFile.addGlobalAttribute ("rows", lines);
    int samples = radiometer.getSampleCount();
    ncFile.addGlobalAttribute ("cols", samples);
    ncFile.addGlobalAttribute ("history", history);

    // Create various data arrays
    // --------------------------
    floatData = new float[samples];
    shortData = new short[samples];
    count = new int[] {1, samples};
    start = new int[] {0, 0};

  } // openOutput

  ////////////////////////////////////////////////////////////

  protected void closeOutput () throws IOException {

    ncFile.flush();
    ncFile.close();

  } // closeOutput

  ////////////////////////////////////////////////////////////

} // CWNCDataWriter class

////////////////////////////////////////////////////////////////////////
