////////////////////////////////////////////////////////////////////////
/*
     FILE: LineScannerLocator.java
  PURPOSE: Computes line scanner earth locations.
   AUTHOR: Peter Hollemans
     DATE: 2007/02/09
  CHANGES: 2007/03/05, PFH, modified to work with new Spacecraft class

  Copyright (C) 2007, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.instrument;

// Imports
// -------
import java.io.*;
import java.util.*;
import terrenus.orbit.*;

/**
 * The <code>LineScannerLocator</code> class computes
 * image scanner earth locations for a cross-track line scanner
 * similar to NOAA AVHRR.
 */
public abstract class LineScannerLocator {

  // Variables
  // ---------

  /** The spacecraft for orbital computations. */
  protected Spacecraft spacecraft;

  /** The ellipsoid model. */
  protected EllipsoidModel earthModel;

  /** The sensor model. */
  protected SensorModel sensorModel;

  /** The line scanner stepping angle in radians. */
  protected double scanStep;

  ////////////////////////////////////////////////////////////

  /**
   * Gets the date for the specified scan line.
   * 
   * @param line the scan line to query.
   *
   * @return the date for the scan line.
   */
  protected abstract Date getLineDate (
    int line
  );

  ////////////////////////////////////////////////////////////

  /**
   * Gets the view vector from a spacecraft position to its geodetic
   * subpoint.  The geodetic subpoint is the point on the reference
   * ellipsoid at which a vector between the ellipsoid surface and the
   * spacecraft is normal to the surface.  This is the standard view
   * vector for the AVHRR instrument on NOAA spacecrafts.
   *
   * @param satPosition the spacecraft position as [x,y,z] in
   * radians.
   *
   * @return the view vector from the spacecraft to the subpoint
   * as an [x,y,z] unit vector.
   */
  public double[] getViewVector (
    double[] satPosition
  ) {

    // Get x-y distance of spacecraft
    // -----------------------------
    double satX = satPosition[VectorOp.X];
    double satY = satPosition[VectorOp.Y];
    double satZ = satPosition[VectorOp.Z];
    double dist = Math.sqrt (satX*satX + satY*satY);

    // Compute simple vector at pole or equator
    // ----------------------------------------
    double[] viewVector;
    if (dist == 0 || satZ == 0)
      viewVector = VectorOp.multiply (satPosition, -1, null);

    // Compute geodetic subpoint and vector
    // ------------------------------------
    else {
      double satGCLat = Math.atan (satZ/dist);
      double radius = VectorOp.magnitude (satPosition);
      double satGDLat = earthModel.GCToGDLat2 (satGCLat, radius);
      double[] satGD = {satGDLat, Math.atan2 (satY, satX)};
      double[] subpoint = earthModel.GDToECEF (satGD, 0, null);
      viewVector = VectorOp.subtract (subpoint, satPosition, null);
    } // else

    VectorOp.normalize (viewVector, viewVector);
    return (viewVector);

  } // getViewVector

  ////////////////////////////////////////////////////////////

  /**
   * Computes the scanner location and angle data for the specified
   * scan line.  Locations are geodetic latitude and longitude.  All
   * angles are in radians.
   *
   * @param line the desired scan line.
   * @param samples the samples to generate locations and angles for,
   * starting at 0.
   * @param latitude the output latitude array.
   * @param longitude the output longitude array.
   * @param satZenith the output spacecraft zenith array.
   * @param solZenith the output solar zenith array.
   * @param relAzimuth the output relative azimuth array.
   */
  public void getLocations (
    int line,
    int[] samples,
    double[] latitude,
    double[] longitude,
    double[] satZenith,
    double[] solZenith,
    double[] relAzimuth
  ) {

    // Get nadir scan time
    // -------------------
    Date date = getLineDate (line);

    // Get spacecraft position and velocity
    // ------------------------------------
    double[] satPosition = spacecraft.getPosition (date, null);
    double radiansPerKm = 1000.0/earthModel.getEquator();
    VectorOp.multiply (satPosition, radiansPerKm, satPosition);
    double[] satVelocity = spacecraft.getVelocity (date, null);

    // Get sun position
    // ----------------
    Sun.setDate (date);
    double[] sunPosition = Sun.getPosition (null);
    VectorOp.multiply (sunPosition, radiansPerKm, sunPosition);

    // Modify positions and velocities for Greenwich hour
    // --------------------------------------------------
    double hourAngle = spacecraft.getHourAngle (date);
    double[][] hourRotation = VectorOp.rotationMatrix (VectorOp.Z, -hourAngle,
      null);
    satPosition = VectorOp.multiply (hourRotation, satPosition, null);
    sunPosition = VectorOp.multiply (hourRotation, sunPosition, null);
    satVelocity = VectorOp.multiply (hourRotation, satVelocity, null);

    // Compute spacecraft axes
    // ----------------------
    double[] viewAxis = getViewVector (satPosition);
    double[] tiltAxis = VectorOp.cross (satVelocity, viewAxis, null);
    double[] rotationAxis = VectorOp.cross (viewAxis, tiltAxis, null);
    VectorOp.normalize (tiltAxis, tiltAxis);
    VectorOp.normalize (rotationAxis, rotationAxis);

    // Setup sensor model
    // ------------------
    sensorModel.setLocation (satPosition);
    sensorModel.setNominalAxes (viewAxis, tiltAxis, rotationAxis); 

    // Create computation vectors
    // --------------------------
    double[] xyzLoc = new double[3];
    double[] geoLoc = new double[2];
    double[] normal = new double[3];
    double[] sunVector = new double[3];
    double[] satVector = new double[3];
    double[] sunInPlane = new double[3];
    double[] satInPlane = new double[3];
    double[] tempVector = new double[3];

    // Loop over each sample location
    // ------------------------------
    for (int i = 0; i < samples.length; i++) {

      // Compute lat/lon location
      // ------------------------
      double rotation = (samples[i]-1023.5)*scanStep;
      sensorModel.getViewLocation (0, rotation, xyzLoc);
      earthModel.ECEFToGD (xyzLoc, 0, geoLoc);
      latitude[i] = geoLoc[0];
      longitude[i] = geoLoc[1];

      // Compute solar zenith angle
      // --------------------------
      earthModel.getNormal (xyzLoc, normal);
      VectorOp.subtract (sunPosition, xyzLoc, sunVector);
      VectorOp.normalize (sunVector, sunVector);
      solZenith[i] = Math.acos (VectorOp.dot (normal, sunVector));

      // Compute spacecraft zenith angle
      // -------------------------------
      VectorOp.subtract (satPosition, xyzLoc, satVector);
      VectorOp.normalize (satVector, satVector);
      satZenith[i] = Math.acos (VectorOp.dot (normal, satVector));
      
      // Compute relative azimuth
      // ------------------------
      VectorOp.subtract (sunVector, VectorOp.multiply (normal, 
        VectorOp.dot (sunVector, normal), tempVector), sunInPlane);
      VectorOp.normalize (sunInPlane, sunInPlane);
      VectorOp.subtract (satVector, VectorOp.multiply (normal,
        VectorOp.dot (satVector, normal), tempVector), satInPlane);      
      VectorOp.normalize (satInPlane, satInPlane);
      relAzimuth[i] = Math.abs (Math.acos (VectorOp.dot (sunInPlane, 
        satInPlane)));

    } // for

  } // getLocations

  ////////////////////////////////////////////////////////////

} // LineScannerLocator class

////////////////////////////////////////////////////////////////////////
