////////////////////////////////////////////////////////////////////////
/*
     FILE: MVISRCalibrator.java
  PURPOSE: Computes MVISR calibration coefficients.
   AUTHOR: Peter Hollemans
     DATE: 2007/02/08
  CHANGES: 2007/03/16, PFH, redesigned class (was MVISRCalibrationGenerator)
           2010/12/18, PFH, removed unneeded cast to CHRPTStream

  Copyright (C) 2007-2010, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.instrument;

// Imports
// -------
import java.io.*;
import java.util.*;
import java.nio.*;
import terrenus.telemetry.*;
import terrenus.orbit.*;

/**
 * The <code>MVISRCalibrator</code> class uses a {@link
 * CHRPTDecoder} to compute MVISR calibration coefficients for a
 * CMA CHRPT data file.  A new set of calibration coefficients
 * may be generated for a user-specified number of consecutive
 * MVISR scan lines.  Calibration data is generated using the
 * methods described in the NOAA KLM User's Guide,
 * http://www2.ncdc.noaa.gov/docs/klm/index.htm.<p>
 *
 * The <code>Record</code> class holds calibration data for one set of
 * consecutive MVISR scan lines.  For visible channels 1, 2, 6, 7, 8,
 * 9, and 10, five coefficients are supplied as: [slope, intercept].
 * Visible albedo in percent may be computed from count value using
 * the equation: albedo = count*slope + intercept.  For thermal
 * channels 3, 4, and 5, three coefficients are supplied as follows:
 * [a0, a1, a2].  Thermal radiance in mW/(m2.sr.cm-1) may be computed
 * from count value using the equation: radiance = a0 + a1*count +
 * a2*count^2.<p>
 */
public class MVISRCalibrator
  extends AbstractRadiometerCalibrator {

  // Constants
  // ---------

  /** First radiation constant (mW/(m2.sr.cm-4)). */
  private static final double C1 = 1.1910427e-5;

  /** Second radiation constant (cm.K). */
  private static final double C2 = 1.4387752;

  /** The offset in the telemetry of PRT reading number 1. */
  private static final int PRT_READING_1 = 16;

  /** The offset in the telemetry of PRT reading number 2. */
  private static final int PRT_READING_2 = 17;

  /** The offset in the telemetry of PRT reading number 3. */
  private static final int PRT_READING_3 = 18;

  /** The offset in the telemetry of PRT reading number 4. */
  private static final int PRT_READING_4 = 19;

  /** The index of channel 3 in various data arrays. */
  private static final int CH3 = 0;

  /** The index of channel 4 in various data arrays. */
  private static final int CH4 = 1;

  /** The index of channel 5 in various data arrays. */
  private static final int CH5 = 2;

  // Variables
  // ---------

  /** The PRT count to temperature coefficients. */
  private double[] prt1_d;
  private double[] prt2_d;
  private double[] prt3_d;
  private double[] prt4_d;

  /** The channel 3 temperature to radiance coefficients. */
  private double ch3_C1_vc3;
  private double ch3_C2_vc;
  private double ch3_vc;
  private double ch3_A;
  private double ch3_B;

  /** The channel 3 nonlinear radiance coefficients. */
  private double ch3_Ns;
  private double ch3_b0;
  private double ch3_b1;
  private double ch3_b2;

  /** The channel 4 temperature to radiance coefficients. */
  private double ch4_C1_vc3;
  private double ch4_C2_vc;
  private double ch4_vc;
  private double ch4_A;
  private double ch4_B;

  /** The channel 4 nonlinear radiance coefficients. */
  private double ch4_Ns;
  private double ch4_b0;
  private double ch4_b1;
  private double ch4_b2;

  /** The channel 5 temperature to radiance coefficients. */
  private double ch5_C1_vc3;
  private double ch5_C2_vc;
  private double ch5_vc;
  private double ch5_A;
  private double ch5_B;

  /** The channel 5 nonlinear radiance coefficients. */
  private double ch5_Ns;
  private double ch5_b0;
  private double ch5_b1;
  private double ch5_b2;

  /** The CHRPT decoder to use for generating calibration. */
  private CHRPTDecoder decoder;

  ////////////////////////////////////////////////////////////

  /** 
   * Creates a new record with visible coefficients filled in.
   *
   * @param startLine the starting scan line for calibration
   * @param endLine the ending scan line for calibration
   */
  private Record createRecord (
    int startLine,
    int endLine
  ) {

    // Create and read visible coefficients
    // ------------------------------------
    Spacecraft craft = decoder.getSpacecraft();

    double[] ch1 = new double[2];
    ch1[0] = Double.parseDouble (craft.getProperty ("mvisr.ch1.slope"));
    ch1[1] = Double.parseDouble (craft.getProperty ("mvisr.ch1.inter"));

    double[] ch2 = new double[2];
    ch2[0] = Double.parseDouble (craft.getProperty ("mvisr.ch2.slope"));
    ch2[1] = Double.parseDouble (craft.getProperty ("mvisr.ch2.inter"));

    double[] ch6 = new double[2];
    ch6[0] = Double.parseDouble (craft.getProperty ("mvisr.ch6.slope"));
    ch6[1] = Double.parseDouble (craft.getProperty ("mvisr.ch6.inter"));

    double[] ch7 = new double[2];
    ch7[0] = Double.parseDouble (craft.getProperty ("mvisr.ch7.slope"));
    ch7[1] = Double.parseDouble (craft.getProperty ("mvisr.ch7.inter"));

    double[] ch8 = new double[2];
    ch8[0] = Double.parseDouble (craft.getProperty ("mvisr.ch8.slope"));
    ch8[1] = Double.parseDouble (craft.getProperty ("mvisr.ch8.inter"));
    
    double[] ch9 = new double[2];
    ch9[0] = Double.parseDouble (craft.getProperty ("mvisr.ch9.slope"));
    ch9[1] = Double.parseDouble (craft.getProperty ("mvisr.ch9.inter"));
    
    double[] ch10 = new double[2];
    ch10[0] = Double.parseDouble (craft.getProperty ("mvisr.ch10.slope"));
    ch10[1] = Double.parseDouble (craft.getProperty ("mvisr.ch10.inter"));

    // Create empty thermal coefficient arrays
    // ---------------------------------------
    double[] ch3 = new double[3];
    double[] ch4 = new double[3];
    double[] ch5 = new double[3];
    
    // Create record
    // -------------
    double[][] coefs = new double[][] {ch1, ch2, ch3, ch4, ch5, ch6, ch7, 
      ch8, ch9, ch10};
    Record record = new Record (coefs, startLine, endLine);

    return (record);

  } // createRecord

  ////////////////////////////////////////////////////////////

  public void calibrate (
    int[] countData,
    double[] calData,
    int channel,
    CalibrationType calType,
    Record record
  ) {

    // Get calibration coefficients
    // ----------------------------
    boolean isVisible;
    double c1_vc3 = 0, c2_vc = 0, a = 0, b = 0;
    switch (channel) {
    case 1: case 2: case 6: case 7: case 8: case 9: case 10:
      isVisible = true; 
      break;
    case 3: 
      isVisible = false; 
      c1_vc3 = ch3_C1_vc3;
      c2_vc = ch3_C2_vc;
      a = ch3_A;
      b = ch3_B;
      break;
    case 4: 
      isVisible = false; 
      c1_vc3 = ch4_C1_vc3;
      c2_vc = ch4_C2_vc;
      a = ch4_A;
      b = ch4_B;
      break;
    case 5: 
      isVisible = false; 
      c1_vc3 = ch5_C1_vc3;
      c2_vc = ch5_C2_vc;
      a = ch5_A;
      b = ch5_B;
      break;
    default: throw new IllegalArgumentException ("Invalid channel number: " +
      channel);
    } // switch
    double[] coefs = record.getCoefs (channel);

    // Check calibration type
    // ----------------------
    if (isVisible) {
      switch (calType) {
      case REFLECTANCE: 
      case ALBEDO: 
        break;
      default: 
        throw new IllegalArgumentException ("Invalid calibration type " +
        calType + " for visible channel " + channel);
      } // switch
    } // if
    else {
      switch (calType) {
      case RADIANCE: 
      case CELSIUS: 
      case KELVIN: 
        break;
      default: throw new IllegalArgumentException ("Invalid calibration type" +
        "for thermal channel " + channel);
      } // switch
    } // else
    
    // Calibrate visible data
    // ----------------------
    if (isVisible) {
      double scale = (calType == CalibrationType.ALBEDO ? 1 : 0.01);
      for (int j = 0; j < countData.length; j++) {
        calData[j] = scale*(countData[j]*coefs[0] + coefs[1]);
      } // for
    } // if

    // Calibrate thermal data
    // ----------------------
    else {
      
      // Initially calibrate to radiance
      // -------------------------------
      for (int j = 0; j < countData.length; j++) {
        calData[j] = coefs[0] + countData[j]*coefs[1] + 
          countData[j]*countData[j]*coefs[2];
      } // for

      // Calibrate to blackbody temperature
      // ----------------------------------
      if (calType != CalibrationType.RADIANCE) {
        if (calType == CalibrationType.CELSIUS) a = a + 273.15*b;
        for (int j = 0; j < countData.length; j++) {
          double tstar = c2_vc/Math.log (1 + (c1_vc3/calData[j]));
          calData[j] = (tstar - a)/b;
        } // for
      } // if
      
    } // else

  } // calibrate

  ////////////////////////////////////////////////////////////

  /** 
   * Creates a new calibrator.
   *
   * @param decoder the decoder to use for data calibration.
   */
  public MVISRCalibrator (
    CHRPTDecoder decoder
  ) {

    super (decoder.getStream().getFrameCount(), 5, 10);

    // Initialize
    // ----------
    this.decoder = decoder;

    // Get sensor coefficients
    // -----------------------
    Spacecraft craft = decoder.getSpacecraft();

    prt1_d = new double[5];
    prt1_d[0] = Double.parseDouble (craft.getProperty ("mvisr.prt1.d0"));
    prt1_d[1] = Double.parseDouble (craft.getProperty ("mvisr.prt1.d1"));
    prt1_d[2] = Double.parseDouble (craft.getProperty ("mvisr.prt1.d2"));
    prt1_d[3] = Double.parseDouble (craft.getProperty ("mvisr.prt1.d3"));
    prt1_d[4] = Double.parseDouble (craft.getProperty ("mvisr.prt1.d4"));

    prt2_d = new double[5];
    prt2_d[0] = Double.parseDouble (craft.getProperty ("mvisr.prt2.d0"));
    prt2_d[1] = Double.parseDouble (craft.getProperty ("mvisr.prt2.d1"));
    prt2_d[2] = Double.parseDouble (craft.getProperty ("mvisr.prt2.d2"));
    prt2_d[3] = Double.parseDouble (craft.getProperty ("mvisr.prt2.d3"));
    prt2_d[4] = Double.parseDouble (craft.getProperty ("mvisr.prt2.d4"));

    prt3_d = new double[5];
    prt3_d[0] = Double.parseDouble (craft.getProperty ("mvisr.prt3.d0"));
    prt3_d[1] = Double.parseDouble (craft.getProperty ("mvisr.prt3.d1"));
    prt3_d[2] = Double.parseDouble (craft.getProperty ("mvisr.prt3.d2"));
    prt3_d[3] = Double.parseDouble (craft.getProperty ("mvisr.prt3.d3"));
    prt3_d[4] = Double.parseDouble (craft.getProperty ("mvisr.prt3.d4"));

    prt4_d = new double[5];
    prt4_d[0] = Double.parseDouble (craft.getProperty ("mvisr.prt4.d0"));
    prt4_d[1] = Double.parseDouble (craft.getProperty ("mvisr.prt4.d1"));
    prt4_d[2] = Double.parseDouble (craft.getProperty ("mvisr.prt4.d2"));
    prt4_d[3] = Double.parseDouble (craft.getProperty ("mvisr.prt4.d3"));
    prt4_d[4] = Double.parseDouble (craft.getProperty ("mvisr.prt4.d4"));

    ch3_vc = Double.parseDouble (craft.getProperty ("mvisr.ch3.vc"));
    ch3_C1_vc3 = C1*Math.pow (ch3_vc, 3);
    ch3_C2_vc = C2*ch3_vc;
    ch3_A = Double.parseDouble (craft.getProperty ("mvisr.ch3.A"));
    ch3_B = Double.parseDouble (craft.getProperty ("mvisr.ch3.B"));
    ch3_Ns = Double.parseDouble (craft.getProperty ("mvisr.ch3.Ns"));
    ch3_b0 = Double.parseDouble (craft.getProperty ("mvisr.ch3.b0"));
    ch3_b1 = Double.parseDouble (craft.getProperty ("mvisr.ch3.b1"));
    ch3_b2 = Double.parseDouble (craft.getProperty ("mvisr.ch3.b2"));

    ch4_vc = Double.parseDouble (craft.getProperty ("mvisr.ch4.vc"));
    ch4_C1_vc3 = C1*Math.pow (ch4_vc, 3);
    ch4_C2_vc = C2*ch4_vc;
    ch4_A = Double.parseDouble (craft.getProperty ("mvisr.ch4.A"));
    ch4_B = Double.parseDouble (craft.getProperty ("mvisr.ch4.B"));
    ch4_Ns = Double.parseDouble (craft.getProperty ("mvisr.ch4.Ns"));
    ch4_b0 = Double.parseDouble (craft.getProperty ("mvisr.ch4.b0"));
    ch4_b1 = Double.parseDouble (craft.getProperty ("mvisr.ch4.b1"));
    ch4_b2 = Double.parseDouble (craft.getProperty ("mvisr.ch4.b2"));

    ch5_vc = Double.parseDouble (craft.getProperty ("mvisr.ch5.vc"));
    ch5_C1_vc3 = C1*Math.pow (ch5_vc, 3);
    ch5_C2_vc = C2*ch5_vc;
    ch5_A = Double.parseDouble (craft.getProperty ("mvisr.ch5.A"));
    ch5_B = Double.parseDouble (craft.getProperty ("mvisr.ch5.B"));
    ch5_Ns = Double.parseDouble (craft.getProperty ("mvisr.ch5.Ns"));
    ch5_b0 = Double.parseDouble (craft.getProperty ("mvisr.ch5.b0"));
    ch5_b1 = Double.parseDouble (craft.getProperty ("mvisr.ch5.b1"));
    ch5_b2 = Double.parseDouble (craft.getProperty ("mvisr.ch5.b2"));

  } // MVISRCalibrator constructor

  ////////////////////////////////////////////////////////////

  /**
   * Gets the temperature measured by one of the PRTs given its
   * count value.
   *
   * @param prtCount the PRT count value.
   * @param prtNumber the PRT number in the range [1..4].
   *
   * @return the temperature of the PRT in Celeius.
   */
  private double getPrtTemp (
    short prtCount,
    int prtNumber
  ) {

    // Find correct conversion constants
    // ---------------------------------
    double[] d;
    switch (prtNumber) {
    case 1: d = prt1_d; break;
    case 2: d = prt2_d; break;
    case 3: d = prt3_d; break;
    case 4: d = prt4_d; break;
    default:
    throw new IllegalArgumentException ("Invalid PRT number");
    } // switch

    // Compute temperature
    // -------------------
    double count;
    double temp = d[0];
    count = prtCount;  temp += d[1] * count;
    count *= prtCount; temp += d[2] * count;
    count *= prtCount; temp += d[3] * count;
    count *= prtCount; temp += d[4] * count;

    return (temp);

  } // getPrtTemp

  ////////////////////////////////////////////////////////////

  public Record getCalibration (
    int startLine,
    int endLine
  ) throws IOException {

    /**
     * The general algorithm that we are performing here is as
     * follows.  For each set of scan lines, we need to:
     *
     * - Gather set of 4 PRT count data values per line
     *
     * - Convert PRT counts to temperature for each PRT
     *
     * - Compute internal BB temperature as mean of all PRT
     *   temperatures
     * 
     * - Compute channel radiance of BB from temperature using channel
     *   3/4/5 coefficients vc, A, B: N3_BB, N4_BB, N5_BB
     * 
     * - For each channel 3/4/5, gather 6 BB count values per scan
     *   line and compute averages: C3_BB, C4_BB, C5_BB
     *
     * - For each channel 3/4/5, gather 10 space count values per scan
     *   line and compute averages: C3_S, C4_S, C5_S
     *
     * - For each channel 3/4/5, compute calibration from BB radiance,
     *   BB counts, space counts, and non-linear correction terms Ns,
     *   b0, b1, b2
     */
    
    // Find usable calibration lines
    // -----------------------------
    int lines = endLine - startLine + 1;
    boolean[] isUsable = new boolean[lines];
    int usableCount = 0;
    for (int i = startLine; i <= endLine; i++) {
      int index = i-startLine;
      isUsable[index] = !decoder.hasErrors (i);
      if (isUsable[index]) usableCount++;
    } // for
    if (usableCount == 0) return (null);

    // Get PRT temperature data
    // ------------------------
    CHRPTStream stream = decoder.getStream();
    double[] prtTempData = new double[lines];
    int prtReadings = 0;
    for (int i = 0; i < lines; i++) {

      // Skip unusable scan line
      // -----------------------
      if (!isUsable[i]) continue;
    
      // Get each PRT temp
      // -----------------
      ShortBuffer telemetry = stream.getBlock (i+startLine, 
        CHRPTStream.TELEMETRY_BLOCK).asShortBuffer();
      double[] prtTemps = new double[] {
        getPrtTemp (telemetry.get (PRT_READING_1), 1),
        getPrtTemp (telemetry.get (PRT_READING_2), 2),
        getPrtTemp (telemetry.get (PRT_READING_3), 3),
        getPrtTemp (telemetry.get (PRT_READING_4), 4)
      };

      // Compute average PRT temperature
      // -------------------------------
      double temp = 0;
      for (int j = 0; j < prtTemps.length; j++)
        temp += prtTemps[j];
      temp /= prtTemps.length;

      // Add temperature to readings
      // ---------------------------
      prtTempData[prtReadings] = temp;
      prtReadings++;

    } // for

    // Check for valid PRT readings
    // ----------------------------
    if (prtReadings == 0) return (null);
        
    // Compute blackbody temperature
    // -----------------------------
    double bbTemp = 0;
    for (int i = 0; i < prtReadings; i++) {
      bbTemp += prtTempData[i];
    } // for
    bbTemp = bbTemp/prtReadings;

    // Compute blackbody radiances
    // ---------------------------
    double[] bbRad = new double[3];
    bbRad[CH3] = ch3_C1_vc3/(Math.exp (ch3_C2_vc/(ch3_A + ch3_B*bbTemp))-1);
    bbRad[CH4] = ch4_C1_vc3/(Math.exp (ch4_C2_vc/(ch4_A + ch4_B*bbTemp))-1);
    bbRad[CH5] = ch5_C1_vc3/(Math.exp (ch5_C2_vc/(ch5_A + ch5_B*bbTemp))-1);

    // Get channel count data
    // ----------------------
    double[] spaceCount = new double[3];
    double[] bbCount = new double[3];
    int channelReadings = 0;
    for (int i = 0; i < lines; i++) {

      // Skip unusable scan line
      // -----------------------
      int scanLine = i+startLine;
      if (!isUsable[i]) continue;

      // Get mean internal target counts
      // -------------------------------
      ShortBuffer target = stream.getBlock (scanLine, 
        CHRPTStream.INTERNAL_TARGET_BLOCK).asShortBuffer();
      for (int channel = CH3; channel <= CH5; channel++) {
        double count = 0;
        for (int j = 0; j < 6; j++) count += target.get (j*10+channel+2);
        bbCount[channel] += count/6;
      } // for

      // Get mean space counts
      // ---------------------
      ShortBuffer space = stream.getBlock (scanLine, 
        CHRPTStream.SPACE_BLOCK).asShortBuffer();
      for (int channel = CH3; channel <= CH5; channel++) {
        double count = 0;
        for (int j = 0; j < 10; j++) count += space.get (j*10+channel+2);
        spaceCount[channel] += count/10;
      } // for

      // Increment number of lines used for channel count data
      // -----------------------------------------------------
      channelReadings++;

    } // for
    if (channelReadings == 0) return (null);
    for (int channel = CH3; channel <= CH5; channel++) {
      bbCount[channel] /= channelReadings;
      spaceCount[channel] /= channelReadings;
    } // for

    // Create calibration record
    // -------------------------
    Record record = createRecord (startLine, endLine);

    double A, B;
    A = (bbRad[CH3] - ch3_Ns)/(bbCount[CH3] - spaceCount[CH3]);
    B = ch3_Ns - A*spaceCount[CH3];
    double[] ch3 = record.getCoefs (3);
    ch3[0] = ch3_b0 + (ch3_b1+1)*B + ch3_b2*B*B;
    ch3[1] = (ch3_b1+1)*A + 2*ch3_b2*A*B;
    ch3[2] = ch3_b2*A*A;

    A = (bbRad[CH4] - ch4_Ns)/(bbCount[CH4] - spaceCount[CH4]);
    B = ch4_Ns - A*spaceCount[CH4];
    double[] ch4 = record.getCoefs (4);
    ch4[0] = ch4_b0 + (ch4_b1+1)*B + ch4_b2*B*B;
    ch4[1] = (ch4_b1+1)*A + 2*ch4_b2*A*B;                              
    ch4[2] = ch4_b2*A*A;

    A = (bbRad[CH5] - ch5_Ns)/(bbCount[CH5] - spaceCount[CH5]);
    B = ch5_Ns - A*spaceCount[CH5];
    double[] ch5 = record.getCoefs (5);
    ch5[0] = ch5_b0 + (ch5_b1+1)*B + ch5_b2*B*B;
    ch5[1] = (ch5_b1+1)*A + 2*ch5_b2*A*B;
    ch5[2] = ch5_b2*A*A;

    return (record);

  } // getCalibration

  ////////////////////////////////////////////////////////////

  /** Tests this class. */
  public static void main (String[] argv) throws Exception {

    // Create stream
    // -------------
    String name = argv[0];
    TelemetryStream stream = TelemetryStreamFactory.create (name);

    // Create calibration records
    // --------------------------
    CHRPTDecoder decoder = (CHRPTDecoder) stream.getDecoder();
    MVISRCalibrator calibrator = new MVISRCalibrator (decoder);
    int lines = stream.getFrameCount();
    for (int i = 0; i < lines; i++) {
      Record record = calibrator.getCalibration (i);
      System.out.println ("Calibration record at line " + i + ":\n" + record);
    } // for

  } // main

  ////////////////////////////////////////////////////////////

} // MVISRCalibrator class

////////////////////////////////////////////////////////////////////////
