////////////////////////////////////////////////////////////////////////
/*
     FILE: MVISRLocator.java
  PURPOSE: Computes MVISR Earth locations.
   AUTHOR: Peter Hollemans
     DATE: 2005/07/20
  CHANGES: 2006/10/14, PFH, added getViewVector() method for computing 
             geodetic subpoint 
           2006/12/04, PFH, changed atan to atan2 in getViewVector()
           2007/01/28, PFH, added use of HRPTReaderFactory
           2007/02/09, PFH, moved main functionality to abstract class
           2007/03/05, PFH, modified to use new Spacecraft class

  Copyright (C) 2005, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.instrument;

// Imports
// -------
import java.io.*;
import java.util.*;
import terrenus.orbit.*;
import terrenus.telemetry.*;

/**
 * The <code>MVISRLocator</code> class uses an {@link
 * CHRPTDecoder} to compute MVISR earth locations for a CMA CHRPT
 * data stream.
 */
public class MVISRLocator
  extends LineScannerLocator {

  // Constants
  // ---------

  /** Standard time unit (STU) period in milliseconds. */
  private static final double STU = 1.001602564;

  /** MVISR nadir scan time in milliseconds. */
  private static final int NADIR_TIME_OFFSET = (int) (34.2*STU);

  // Variables
  // ---------

  /** The CHRPT decoder to use for generating dates for each scan line. */
  private CHRPTDecoder decoder;
  
  ////////////////////////////////////////////////////////////

  protected Date getLineDate (
    int line
  ) {

    return (new Date (decoder.getTime (line) + NADIR_TIME_OFFSET));

  } // getLineDate

  ////////////////////////////////////////////////////////////

  /** 
   * Creates a new locator.
   *
   * @param decoder the decoder to use for time and spacecraft
   * information.
   */
  public MVISRLocator (
    CHRPTDecoder decoder
  ) {

    // Initialize
    // ----------
    this.decoder = decoder;
    spacecraft = decoder.getSpacecraft();
    earthModel = EllipsoidModelFactory.getWGS72();
    sensorModel = new SensorModel (earthModel);

    // Get scan step value
    // -------------------
    scanStep = Math.toRadians (Double.parseDouble (spacecraft.getProperty (
      "mvisr.scan.step")));

  } // MVISRLocator constructor

  ////////////////////////////////////////////////////////////

} // MVISRLocator class

////////////////////////////////////////////////////////////////////////
