////////////////////////////////////////////////////////////////////////
/*
     FILE: AbstractOrbitalElementSource.java
  PURPOSE: Provides two-line orbital elements.
   AUTHOR: Peter Hollemans
     DATE: 2005/08/05
  CHANGES: n/a

  Copyright (C) 2005, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.orbit;

// Imports
// -------
import java.util.*;
import sgp.*;

/**
 * The <code>AbstractOrbitalElementSource</code> class implements
 * useful method for all orbital element sources.
 */
public abstract class AbstractOrbitalElementSource 
  implements OrbitalElementSource {

  // Constants
  // ---------

  /** The UTC timezone. */
  private static final TimeZone UTC = TimeZone.getTimeZone ("GMT+0");

  /** The number of milliseconds per day. */
  private static final int MSEC_PER_DAY = 86400000;

  ////////////////////////////////////////////////////////////

  /**
   * Gets the element epoch date.
   * 
   * @param elements the two-line orbital element strings.
   *
   * @return the element epoch date.
   */
  public static Date getEpoch (
    String[] elements
  ) {

    // Get element epoch
    // -----------------
    TLE tle = new TLE ("UNKNOWN", elements[0], elements[1]);
    MarcaDeTiempo epoch = new MarcaDeTiempo();
    Tiempo.julianTimeToTime (Tiempo.julianDateOfEpoch (tle.getEpoch()), epoch);
    Calendar cal = new GregorianCalendar (UTC);
    cal.set (epoch.yr, epoch.mo, epoch.dy, epoch.hr, epoch.mi, epoch.se);
    cal.set (Calendar.MILLISECOND, epoch.hu);
    
    return (cal.getTime());

  } // getEpoch

  ////////////////////////////////////////////////////////////

  /** 
   * Computes the day offset from the element epoch.
   *
   * @param elements the two-line orbital element strings.
   * @param date the date to compute the offset.
   * 
   * @return the day offset from the element epoch, positive if the
   * date is after the element epoch, and negative if before.
   */
  public static double getEpochOffset (
    String[] elements,
    Date date
  ) {

    Date epoch = getEpoch (elements);
    double days = (double) (date.getTime() - epoch.getTime()) / MSEC_PER_DAY;

    return (days);

  } // getEpochOffset

  ////////////////////////////////////////////////////////////

} // AbstractOrbitalElementSource class

////////////////////////////////////////////////////////////////////////
