////////////////////////////////////////////////////////////////////////
/*
     FILE: OrbitalElementManager.java
  PURPOSE: Provides two-line orbital elements from various sources.
   AUTHOR: Peter Hollemans
     DATE: 2005/10/20
  CHANGES: n/a

  Copyright (C) 2005, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.orbit;

// Imports
// -------
import java.util.*;

/**
 * The <code>OrbitalElementManager</code> uses a list of {@link
 * OrbitalElementSource} objects to provide orbital element data for
 * any number of satellites.  Both current and archived element
 * sources may be registered for the manager to use.
 */
public class OrbitalElementManager
  implements OrbitalElementSource {

  // Variables
  // ---------
  
  /** The list of sources. */
  private List<OrbitalElementSource> sourceList = 
    new ArrayList<OrbitalElementSource>();

  /** The static instance. */
  private static OrbitalElementManager instance;

  ////////////////////////////////////////////////////////////

  /** Gets an instance of the manager. */
  public static OrbitalElementManager getInstance () {

    if (instance == null) instance = new OrbitalElementManager();
    return (instance);

  } // getInstance

  ////////////////////////////////////////////////////////////

  /**
   * Registers a new element source with this manager.
   *
   * @param source the new element source.
   */
  public void registerSource (
    OrbitalElementSource source
  ) {

    sourceList.add (source);

  } // registerSource

  ////////////////////////////////////////////////////////////

  /**
   * Gets the orbital elements for the specified satellite and date.
   * The first source to report back a set of elements that match the
   * satellite and satisfy the time difference tolerance is used.
   * 
   * @param satid the NORAD orbital element ID for the satellite.
   * @param date the date for orbital element retrieval.
   * @param maxDays the maximum allowed difference between date and orbital
   * element epoch in days.
   *
   * @return the orbital elements as two strings, or null if none could be
   * found within the specified tolerance.
   */
  public String[] getElements (
    int satid,
    Date date,
    double maxDays
  ) {

    // Iterate over each source until found
    // ------------------------------------
    for (OrbitalElementSource source : sourceList) {
      String[] elements = source.getElements (satid, date, maxDays);
      if (elements != null) return (elements);
    } // for

    return (null);

  } // getElements 

  ////////////////////////////////////////////////////////////

} // OrbitalElementManager class

////////////////////////////////////////////////////////////////////////
