////////////////////////////////////////////////////////////////////////
/*
     FILE: Sun.java
  PURPOSE: Provides sun position information.
   AUTHOR: Peter Hollemans
     DATE: 2005/10/23
  CHANGES: n/a

  Copyright (C) 2005, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.orbit;

// Imports
// -------
import java.util.*;
import sgp.*;

/**
 * The <code>Sun</code> class models the sun position over time.  The
 * position is computed relative to a WGS 72 Earth model.
 */
public class Sun {

  // Constants
  // ---------

  /** The UTC timezone. */
  private static final TimeZone UTC = TimeZone.getTimeZone ("GMT+0");

  ////////////////////////////////////////////////////////////

  /** 
   * Sets the date object and performs the sun position computation.
   * This method must be called prior to using {@link #getPosition}.
   *
   * @param date the date for the sun computations.
   */
  public static void setDate (
    Date date
  ) {

    // Compute new position
    // --------------------
    Calendar targetDate = new GregorianCalendar (UTC);
    targetDate.setTime (date);
    MarcaDeTiempo targetTime = new MarcaDeTiempo();
    targetTime.setValues (
      targetDate.get (Calendar.YEAR),
      targetDate.get (Calendar.MONTH),
      targetDate.get (Calendar.DATE),
      targetDate.get (Calendar.HOUR_OF_DAY),
      targetDate.get (Calendar.MINUTE),
      targetDate.get (Calendar.SECOND),
      targetDate.get (Calendar.MILLISECOND)
    );
    Sol.calcularPosicion (targetTime);
    
  } // setDate

  ////////////////////////////////////////////////////////////

  /**
   * Gets the position of the sun in Earth-centered inertial
   * coordinates.
   * 
   * @param output the output position array to fill.  If null, an
   * output array is created.
   *
   * @return the output position array as [x,y,z] in kilometers.
   */
  public static double[] getPosition (
    double[] output
  ) {
  
    if (output == null) output = new double[3];
    for (int i = 0; i < 3; i++) output[i] = Sol.pos[i];
    return (output);

  } // getPosition

  ////////////////////////////////////////////////////////////

} // Sun class

////////////////////////////////////////////////////////////////////////
