////////////////////////////////////////////////////////////////////////
/*
     FILE: VectorOp.java
  PURPOSE: Performs vector and matrix operations.
   AUTHOR: Peter Hollemans
     DATE: 2005/10/14
  CHANGES: n/a

  Copyright (C) 2005, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.orbit;

/**
 * The <code>VectorOp</code> class performs various vector and matrix
 * operations.
 */
public class VectorOp {

  // Constants
  // ---------
 
  /** The index of the x coordinate. */
  public static final int X = 0;

  /** The index of the y coordinate. */
  public static final int Y = 1;

  /** The index of the z coordinate. */
  public static final int Z = 2;

  ////////////////////////////////////////////////////////////

  /**
   * Multiplies a vector by a scalar.
   * 
   * @param vector the vector to multiply.
   * @param factor the scalar multiple.
   * @param output the output vector to fill.  If null, an
   * output vector is created.
   *
   * @return the output vector.
   */
  public static double[] multiply (
    double[] vector,
    double factor,
    double[] output
  ) {

    if (output == null) output = new double[3];
    for (int i = 0; i < 3; i++)
      output[i] = vector[i]*factor;
    return (output);    

  } // multiply

  ////////////////////////////////////////////////////////////

  /**
   * Computes the vector 2-norm magnitude.
   * 
   * @param vector the vector to compute magnitude.
   *
   * @return the vector 2-norm magnitude.
   */
  public static double magnitude (
    double[] vector
  ) {

    return (Math.sqrt (dot (vector, vector)));

  } // magnitude

  ////////////////////////////////////////////////////////////

  /**
   * Normalizes a vector to unit magnitude.
   * 
   * @param vector the vector to normalize.
   * @param output the output vector to fill.  If null, an
   * output vector is created.
   *
   * @return the output vector.
   */
  public static double[] normalize (
    double[] vector,
    double[] output
  ) {

    double mag = magnitude (vector);
    return (multiply (vector, 1/mag, output));

  } // normalize

  ////////////////////////////////////////////////////////////

  /**
   * Computes the cross product of two vectors, a x b.
   * 
   * @param vectorA the first vector in the product.
   * @param vectorB the second vector in the product.
   * @param output the output vector to fill.  If null, an
   * output vector is created.
   *
   * @return the output product vector.
   */
  public static double[] cross (
    double[] vectorA,
    double[] vectorB,
    double[] output
  ) {

    if (output == null) output = new double[3];
    output[X] = vectorA[Y]*vectorB[Z] - vectorA[Z]*vectorB[Y];
    output[Y] = vectorA[Z]*vectorB[X] - vectorA[X]*vectorB[Z];
    output[Z] = vectorA[X]*vectorB[Y] - vectorA[Y]*vectorB[X];
    return (output);

  } // cross

  ////////////////////////////////////////////////////////////

  /**
   * Computes the sum two vectors, a + b.
   * 
   * @param vectorA the first vector in the sum.
   * @param vectorB the second vector in the sum.
   * @param output the output vector to fill.  If null, an
   * output vector is created.
   *
   * @return the output sum vector.
   */
  public static double[] add (
    double[] vectorA,
    double[] vectorB,
    double[] output
  ) {

    if (output == null) output = new double[3];
    for (int i = 0; i < 3; i++)
      output[i] = vectorA[i] + vectorB[i];
    return (output);    

  } // add

  ////////////////////////////////////////////////////////////

  /**
   * Computes the difference of two vectors, a - b.
   * 
   * @param vectorA the first vector in the difference.
   * @param vectorB the second vector in the difference.
   * @param output the output vector to fill.  If null, an
   * output vector is created.
   *
   * @return the output difference vector.
   */
  public static double[] subtract (
    double[] vectorA,
    double[] vectorB,
    double[] output
  ) {

    if (output == null) output = new double[3];
    for (int i = 0; i < 3; i++)
      output[i] = vectorA[i] - vectorB[i];
    return (output);    

  } // subtract

  ////////////////////////////////////////////////////////////

  /**
   * Applies a matrix transform to a vector.
   * 
   * @param matrix the matrix to apply.
   * @param vector the vector to transform.
   * @param output the output vector to fill.  If null, an
   * output vector is created.
   *
   * @return the output transformed vector.
   */
  public static double[] multiply (
    double[][] matrix, 
    double[] vector,
    double[] output
  ) {

    if (output == null) output = new double[3];
    for (int row = 0; row < 3; row++) {
      output[row] = 0;
      for (int col = 0; col < 3; col++)
        output[row] += matrix[row][col] * vector[col];
    } // for
    return (output);

  } // multiply

  ////////////////////////////////////////////////////////////

  /**
   * Computes the dot product of two vectors, a (dot) b.
   * 
   * @param vectorA the first vector in the product.
   * @param vectorB the second vector in the product.
   *
   * @return the output dot product.
   */
  public static double dot (
    double[] vectorA, 
    double[] vectorB
  ) {

    double sum = 0;
    for (int i = 0; i < 3; i++)
      sum += vectorA[i] * vectorB[i];
    return (sum);

  } // dot

  ////////////////////////////////////////////////////////////

  /**
   * Computes a rotation matrix.
   * 
   * @param axis the axis about which the matrix should rotate: X, Y, or Z.
   * @param theta the angle to rotate in radians.  A positive angle
   * will rotate coordinates in a counter-clockwise direction (right
   * hand rule, thumb pointing along the positive axis).
   * @param output the output matrix to fill.  If null, an output
   * matrix is created.
   *
   * @return the output rotation matrix.
   */
  public static double[][] rotationMatrix (
    int axis, 
    double theta,
    double[][] output
  ) {

    if (output == null) output = new double[3][3];
    double sin_theta = Math.sin (-theta);
    double cos_theta = Math.cos (-theta);
    switch (axis) {
    case X:
      output[0][0] = 1; output[0][1] = 0; output[0][2] = 0;
      output[1][0] = 0; output[1][1] = cos_theta; output[1][2] = sin_theta;
      output[2][0] = 0; output[2][1] = -sin_theta; output[2][2] = cos_theta;
      break;
    case Y:
      output[0][0] = cos_theta; output[0][1] = 0; output[0][2] = -sin_theta;
      output[1][0] = 0; output[1][1] = 1; output[1][2] = 0;
      output[2][0] = sin_theta; output[2][1] = 0; output[2][2] = cos_theta;
      break;
    case Z:
      output[0][0] = cos_theta; output[0][1] = sin_theta; output[0][2] = 0;
      output[1][0] = -sin_theta; output[1][1] = cos_theta; output[1][2] = 0;
      output[2][0] = 0; output[2][1] = 0; output[2][2] = 1;
      break;
    } // switch
    return (output);

  } // rotationMatrix

  ////////////////////////////////////////////////////////////

  /**
   * Computes the product of two matrices, AB.
   * 
   * @param matrixA the first matrix in the product.
   * @param matrixB the second matrix in the product.
   * @param output the output matrix to fill.  If null, an
   * output matrix is created.
   *
   * @return the output product matrix.
   */
  public static double[][] multiply (
    double[][] matrixA, 
    double[][] matrixB,
    double[][] output
  ) {

    if (output == null) output = new double[3][3];
    for (int row = 0; row < 3; row++) {
      for (int col = 0; col < 3; col++) {
        output[row][col] = 0;
        for (int i = 0; i < 3; i++)
          output[row][col] += matrixA[row][i] * matrixB[i][col];
      } // for
    } // for
    return (output);

  } // multiply

  ////////////////////////////////////////////////////////////

  /** Converts a vector to a string. */
  public static String toString (
    double[] vector
  ) {

    StringBuffer buffer = new StringBuffer();
    buffer.append ("[");
    for (int i = 0; i < vector.length; i++) {
      buffer.append (Double.toString (vector[i]));
      if (i != vector.length-1) buffer.append (", ");
    } // for
    buffer.append ("]");

    return (buffer.toString());

  } // toString

  ////////////////////////////////////////////////////////////

} // VectorOp class

////////////////////////////////////////////////////////////////////////
