////////////////////////////////////////////////////////////////////////
/*
     FILE: CHRPTStream.java
  PURPOSE: Interface for reading CHRPT data.
   AUTHOR: Peter Hollemans
           Mohammed Kamal
     DATE: 2007/02/07
  CHANGES: 2007/03/16, PFH, modifid to extend TelemetryStream

  Copyright (C) 2007, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.telemetry;

// Imports
// -------
import java.io.*;
import java.nio.*;

/**
 * The <code>CHRPTStream</code> is a telemetry stream of China
 * Meterological Administration (CMA) CHRPT data.  The stream delivers
 * CHRPT data as described by the Dundee Satellite Receiving Station
 * website (http://www.sat.dundee.ac.uk/hrptformat.html) as follows:
 *
 * <table border="1">
 * 
 *   <tr>
 *     <th>Function</th>
 *     <th>Number<br>of words</th>
 *     <th>Word<br>Position</th>
 *     <th>Description</th>
 *   </tr>
 * 
 *   <tr>
 *     <td rowspan="6">Frame sync</td>
 *     <td rowspan="6">6</td>
 *     <td>1</td>
 *     <td>1010000100</td>
 *   </tr>
 * 
 *   <tr>
 *     <td>2</td>
 *     <td>0101101111</td>
 *   </tr>
 * 
 *   <tr>
 *     <td>3</td>
 *     <td>1101011100</td>
 *   </tr>
 * 
 *   <tr>
 *     <td>4</td>
 *     <td>0110011101</td>
 *   </tr>
 * 
 *   <tr>
 *     <td>5</td>
 *     <td>1000001111</td>
 *   </tr>
 *
 *   <tr>
 *     <td>6</td>
 *     <td>0010010101</td>
 *   </tr>
 * 
 *   <tr>
 *     <td rowspan="2">ID</td>
 *     <td rowspan="2">2</td>
 *     <td>7</td>
 *     <td>Bits 1-4 = Satellite ID (FY-1D = 1101, FY-1C = 1100)</td>
 *   </tr>
 * 
 *   <tr>
 *     <td>8</td>
 *     <td>Spare</td>
 *   </tr>
 *
 *   <tr>
 *     <td rowspan="4">Time code</td>
 *     <td rowspan="4">4</td>
 *     <td>9</td>
 *     <td>Bits 1-9 = binary day count (in reality, zero-filled)</td>
 *   </tr>
 * 
 *   <tr>
 *     <td>10</td>
 *     <td>Bits 1-3 = 101, bits 4-10 = part of millisecond of day count</td>
 *   </tr>
 * 
 *   <tr>
 *     <td>11</td>
 *     <td>Part of millisecond</td>
 *   </tr>
 * 
 *   <tr>
 *     <td>12</td>
 *     <td>Remainder of millisecond</td>
 *   </tr>
 * 
 *   <tr>
 *     <td rowspan="8">Telemetry</td>
 *     <td rowspan="8">20</td>
 *     <td>13-22</td>
 *     <td>Ramp calibration of 10 channels of MVISR</td>
 *   </tr>
 * 
 *   <tr>
 *     <td>23</td>
 *     <td>Temperature of first stage of radiant cooler A</td>
 *   </tr>
 * 
 *   <tr>
 *     <td>24</td>
 *     <td>Temperature of first stage of radiant cooler B</td>
 *   </tr>
 * 
 *   <tr>
 *     <td>25</td>
 *     <td>Temperature of second stage of radiant cooler A</td>
 *   </tr>
 * 
 *   <tr>
 *     <td>26</td>
 *     <td>Temperature of second stage of radiant cooler B</td>
 *   </tr>
 * 
 *   <tr>
 *     <td>27</td>
 *     <td>Temp controlling voltage of 2nd stage radiant cooler A</td>
 *   </tr>
 * 
 *   <tr>
 *     <td>28</td>
 *     <td>Temp controlling voltage of 2nd stage radiant cooler B</td>
 *   </tr>
 * 
 *   <tr>
 *     <td>29-32</td>
 *     <td>Temp of sheath of radiometer, 1 word per platinium resistance 
 *       bulb</td>
 *   </tr>
 * 
 *   <tr>
 *     <td>Internal target data</td>
 *     <td>60</td>
 *     <td>33-92</td>
 *     <td>6 sampling words for every channel</td>
 *   </tr>
 * 
 *   <tr>
 *     <td>Space</td>
 *     <td>100</td>
 *     <td>93-192</td>
 *     <td>10 sampling words for every channel</td>
 *   </tr>
 * 
 *   <tr>
 *    <td>Spare</td>
 *    <td>1408</td>
 *     <td>193-1600</td>
 *     <td>
 *       00000 11101<br>
 *       00000 01100<br>
 *       10010 01010<br>
 *       ...<br>
 *       00100 11110<br>
 *       11111 11000<br>
 *       11000 01101<br>
 *       11011 00101
 *     </td>
 *   </tr>
 * 
 *   <tr>
 *     <td>Earth data</td>
 *     <td>20480</td>
 *     <td>1601-22080</td>
 *     <td>
 *       2048 words for every channel from channel 1 to channel 10.
 *       Each frame contains the data obtained during one earth scan
 *       of the MVISR sensor.  Time multiplexed.
 *     </td>
 *   </tr>
 * 
 *   <tr>
 *     <td>Aux. sync</td>
 *     <td>100</td>
 *     <td>22081-22180</td>
 *     <td>
 *       11111 00010<br>
 *       11111 10011<br>
 *       01101 10101<br>
 *       ...<br>
 *       01111 10000<br>
 *       11110 01100
 *     </td>
 *   </tr>
 * 
 * </table>
 * <p>
 *
 * The methods commit the subclass to the delivery of
 * CHRPT minor frames in the same format as outlined in the
 * documentation, regardless of the underlying capture station
 * storage format.  In the case where the storage format does not
 * store some part of the data, the methods may create artificial
 * data (such as the frame synchronization) or may resort to
 * zero-filling if there is no reasonable alternative.  The only
 * modification to the original documented format is that words
 * returned by the methods contain data in the least significant
 * 10 bits of a 16-bit value rather than the original 10-bit
 * word.<p>
 * 
 * Subclasses must implement the {@link #getFormat} method and
 * override any other methods such as {@link #getFramePosition}
 * that would required changes if the file format is structurally
 * very different from a broadcast CHRPT stream.
 */
public abstract class CHRPTStream 
  extends AbstractTelemetryStream {

  // Constants
  // ---------

  /** The array of CHRPT frame block offsets in bytes. */
  public static final int[] BLOCK_OFFSETS = new int[] {
    0*2,     // Frame sync
    6*2,     // ID (MVISR)
    8*2,     // Time code
    12*2,    // Telemetry (MVISR)
    32*2,    // Internal target data (MVISR)
    92*2,    // Space data (MVISR)
    192*2,   // Spare words
    1600*2,  // Earth data (MVISR) 
    22080*2  // Auxiliary sync
  };

  /** The array of CHRPT frame block lengths in bytes. */
  public static final int[] BLOCK_LENGTHS = new int[] {
    6*2,     // Frame sync
    2*2,     // ID (MVISR)
    4*2,     // Time code
    20*2,    // Telemetry (MVISR)
    60*2,    // Internal target data (MVISR)
    100*2,   // Space data (MVISR)
    1408*2,  // Spare words
    20480*2, // Earth data (MVISR)
    100*2    // Auxiliary sync
  };

  /** Frame sync block index. */
  public static final int FRAME_SYNC_BLOCK = 0;

  /** ID block index. */
  public static final int ID_BLOCK = 1;

  /** Time code block index. */
  public static final int TIME_CODE_BLOCK = 2;

  /** Telemetry block index. */
  public static final int TELEMETRY_BLOCK = 3;

  /** Internal target block index. */
  public static final int INTERNAL_TARGET_BLOCK = 4;

  /** Space data block index. */
  public static final int SPACE_BLOCK = 5;

  /** Spare words block index. */
  public static final int SPARE_BLOCK = 6;

  /** Earth data block index. */                                              
  public static final int EARTH_BLOCK = 7;

  /** Auxiliary sync block index. */
  public static final int AUX_SYNC_BLOCK = 8;

  ////////////////////////////////////////////////////////////

  public int getBlockCount() { return (BLOCK_OFFSETS.length); }

  ////////////////////////////////////////////////////////////

  protected int getBlockPosition (int block) { return (BLOCK_OFFSETS[block]); }

  ////////////////////////////////////////////////////////////

  protected int getBlockLength (int block) { return (BLOCK_LENGTHS[block]); }

  ////////////////////////////////////////////////////////////

  public TelemetryDecoder getDecoder() throws IOException { 

    return (new CHRPTDecoder (this)); 

  } // getDecoder

  ////////////////////////////////////////////////////////////

  protected boolean checkFrame (
    int frame
  ) {

    ByteBuffer buffer = getBlock (frame, FRAME_SYNC_BLOCK);
    int diff = HRPTFrameSyncChecker.getInstance().getDifferenceCount (buffer);
    if (diff == 0) return (true);
    else return (false);

  } // checkFrame

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the day count for the specified frame. The day count is not
   * available from the CHRPT telemetry stream from the spacecraft
   * because the bits in the time code where the day count are
   * supposed to be are zero filled.
   */
  public abstract int getDayCount (int frame) throws IOException;

  ////////////////////////////////////////////////////////////

} // CHRPTStream class

////////////////////////////////////////////////////////////////////////
