////////////////////////////////////////////////////////////////////////
/*
     FILE: CMA1A5CHRPTStream.java
  PURPOSE: Reads CMA 1A5 format CHRPT data.
   AUTHOR: Peter Hollemans
     DATE: 2007/04/17
  CHANGES: n/a

  Copyright (C) 2007, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.telemetry;

// Imports
// -------
import java.io.*;
import java.nio.*;

/**
 * The <code>CMA1A5CHRPTStream</code> class reads China Meteorological
 * Administration (CMA) 1A5 format CHRPT data:
 * <ul>
 *   <li>44360 byte header</li>
 *   <li>44360 bytes per CHRPT frame composed of 3200 byte frame
 *   header and 40960 byte channel data as 20480 16-bit little endian
 *   values</li>
 * </ul><p>
 *
 * The 44360 byte header contains the following values in little endian
 * byte order:
 * <table border="1">
 *
 *   <tr>
 *     <th>Byte offset</th>
 *     <th>Java type</th>
 *     <th>Description</th>
 *   </tr>
 *
 *   <tr>
 *     <td>0</td>
 *     <td>short</td>
 *     <td>Satellite ID</td>
 *   </tr>
 *
 *   <tr>
 *     <td>2</td>
 *     <td>short</td>
 *     <td>Last two digits of starting year</td>
 *   </tr>
 *
 *   <tr>
 *     <td>4</td>
 *     <td>int</td>
 *     <td>Starting millisecond</td>
 *   </tr>
 *
 *   <tr>
 *     <td>8</td>
 *     <td>short</td>
 *     <td>Starting day</td>
 *   </tr>
 *
 *   <tr>
 *     <td>10</td>
 *     <td>short</td>
 *     <td>Last two digits of ending year</td>
 *   </tr>
 *
 *   <tr>
 *     <td>12</td>
 *     <td>int</td>
 *     <td>Ending millisecond</td>
 *   </tr>
 *
 *   <tr>
 *     <td>16</td>
 *     <td>short</td>
 *     <td>Ending day</td>
 *   </tr>
 *
 *   <tr>
 *     <td>18</td>
 *     <td>short</td>
 *     <td>Good scan line count</td>
 *   </tr>
 *
 *   <tr>
 *     <td>20</td>
 *     <td>short</td>
 *     <td>Total record count (scan line count plus header)</td>
 *   </tr>
 *
 *   <tr>
 *     <td>22</td>
 *     <td>short</td>
 *     <td>Number of errors</td>
 *   </tr>
 *
 *   <tr>
 *     <td>24</td>
 *     <td>short</td>
 *     <td>Error ratio</td>
 *   </tr>
 *
 *   <tr>
 *     <td>26</td>
 *     <td>short</td>
 *     <td>Reserved block #1</td>
 *   </tr>
 *
 *   <tr>
 *     <td>28</td>
 *     <td>short</td>
 *     <td>Time error</td>
 *   </tr>
 *
 *   <tr>
 *     <td>30</td>
 *     <td>short</td>
 *     <td>Missing scan lines</td>
 *   </tr>
 *
 *   <tr>
 *     <td>32</td>
 *     <td>short</td>
 *     <td>Ramp slope analysis result</td>
 *   </tr>
 *
 *   <tr>
 *     <td>34</td>
 *     <td>short</td>
 *     <td>Reserved block #2</td>
 *   </tr>
 *
 *   <tr>
 *     <td>36</td>
 *     <td>float[40]</td>
 *     <td>Calibration coefficients for 10 channels:
 *     <ul>
 *       <li>Slope
 *       <li>Intercept
 *       <li>Standard deviation in slope
 *       <li>Standard deviation in intercept
 *     </ul></td>
 *   </tr>
 *
 *   <tr>
 *     <td>196</td>
 *     <td>short</td>
 *     <td>Reserved block #3</td>
 *   </tr>
 *
 *   <tr>
 *     <td>198</td>
 *     <td>short</td>
 *     <td>Orbit number</td>
 *   </tr>
 *
 *   <tr>
 *     <td>200</td>
 *     <td>double[7]</td>
 *     <td>Orbital element values:
 *     <ul>
 *       <li>Epoch time
 *       <li>Orbital semi-major axis
 *       <li>Orbital ecentricity
 *       <li>Orbital inclination
 *       <li>Right ascension of ascending node
 *       <li>Argument of perigee
 *       <li>???
 *     </ul></td>
 *   </tr>
 *
 *   <tr>
 *     <td>256</td>
 *     <td>short</td>
 *     <td>Direction (0 = descending, 1 = ascending)</td>
 *   </tr>
 *
 *   <tr>
 *     <td>258</td>
 *     <td>short</td>
 *     <td>Orbit type</td>
 *   </tr>
 *
 *   <tr>
 *     <td>260</td>
 *     <td>short</td>
 *     <td>Start orbit number</td>
 *   </tr>
 *
 *   <tr>
 *     <td>262</td>
 *     <td>short[5]</td>
 *     <td>Reserved block #4</td>
 *   </tr>
 *
 *   <tr>
 *     <td>272</td>
 *     <td>double[3]</td>
 *     <td>Satellite attitude: roll, pitch, yaw</td>
 *   </tr>
 *
 *   <tr>
 *     <td>296</td>
 *     <td>short[5]</td>
 *     <td>Reserved block #5</td>
 *   </tr>
 *
 *   <tr>
 *     <td>306</td>
 *     <td>float[8]</td>
 *     <td>Latitude and longitude of 4 corner points</td>
 *   </tr>
 *
 *   <tr>
 *     <td>338</td>
 *     <td>double</td>
 *     <td>Start time in seconds since Jan 1 1980 00:00 UTC</td>
 *   </tr>
 *
 *   <tr>
 *     <td>346</td>
 *     <td>double</td>
 *     <td>End time in seconds since Jan 1 1980 00:00 UTC</td>
 *   </tr>
 *
 *   <tr>
 *     <td>354</td>
 *     <td>short</td>
 *     <td>Grid flag</td>
 *   </tr>
 *
 *   <tr>
 *     <td>356</td>
 *     <td>short</td>
 *     <td>Grid starting latitude</td>
 *   </tr>
 *
 *   <tr>
 *     <td>358</td>
 *     <td>short</td>
 *     <td>Grid starting longitude</td>
 *   </tr>
 *
 *   <tr>
 *     <td>360</td>
 *     <td>short[22000]</td>
 *     <td>Reserved block #6</td>
 *   </tr>
 *
 * </table><p>
 *
 * Each CHRPT frame header contains the following values in little
 * endian byte order:
 * <table border="1">
 *
 *   <tr>
 *     <th>Byte offset</th>
 *     <th>Java type</th>
 *     <th>Description</th>
 *   </tr>
 *
 *   <tr>
 *     <td>0</td>
 *     <td>short</td>
 *     <td>Scan line number</td>
 *   </tr>
 *
 *   <tr>
 *     <td>2</td>
 *     <td>short</td>
 *     <td>Year</td>
 *   </tr>
 *
 *   <tr>
 *     <td>4</td>
 *     <td>int</td>
 *     <td>Millisecond</td>
 *   </tr>
 *
 *   <tr>
 *     <td>8</td>
 *     <td>short</td>
 *     <td>Day</td>
 *   </tr>
 *
 *   <tr>
 *     <td>10</td>
 *     <td>short</td>
 *     <td>Reserved block #1</td>
 *   </tr>
 *
 *   <tr>
 *     <td>12</td>
 *     <td>short</td>
 *     <td>Quality flag</td>
 *   </tr>
 *
 *   <tr>
 *     <td>14</td>
 *     <td>short</td>
 *     <td>Reserved block #2</td>
 *   </tr>
 *
 *   <tr>
 *     <td>16</td>
 *     <td>float[20]</td>
 *     <td>Slope and intercept for 10 channels</td>
 *   </tr>
 *
 *   <tr>
 *     <td>96</td>
 *     <td>float[51]</td>
 *     <td>Solar zenith angle for 51 points in degrees</td>
 *   </tr>
 *
 *   <tr>
 *     <td>300</td>
 *     <td>float[102]</td>
 *     <td>Latitude and longitude for 51 points in degrees</td>
 *   </tr>
 *
 *   <tr>
 *     <td>708</td>
 *     <td>short[193]</td>
 *     <td>First 193 words of CHRPT frame data</td>
 *   </tr>
 *
 *   <tr>
 *     <td>1094</td>
 *     <td>short</td>
 *     <td>???</td>
 *   </tr>
 *
 *   <tr>
 *     <td>1096</td>
 *     <td>float[51]</td>
 *     <td>Satellite zenith angle for 51 points in degrees</td>
 *   </tr>
 *
 *   <tr>
 *     <td>1300</td>
 *     <td>float[51]</td>
 *     <td>Relative azimuth angle for 51 points in degrees</td>
 *   </tr>
 *
 *   <tr>
 *     <td>1504</td>
 *     <td>short[848]</td>
 *     <td>Reserved block #3</td>
 *   </tr>
 *
 * </table>
 */
public class CMA1A5CHRPTStream 
  extends CHRPTStream {

  // Constants
  // ---------

  /** Short word offset into file header for start year. */
  private static final int YEAR_OFFSET = 1;

  /** Byte offset into line header for day count. */
  private static final int DAY_OFFSET = 8;

  /** Byte offset into line header for raw CHRPT data. */
  private static final int CHRPT_OFFSET = 708;

  /** The minor frame length in bytes. */
  private static final int FRAME_LENGTH = 44360;

  /** The earth data offset in bytes. */
  private static final int EARTH_DATA_OFFSET = 3200 - CHRPT_OFFSET;

  // Variables
  // ---------

  /** The fake buffer of 1408 spare words. */
  private ByteBuffer spareBuffer;

  /** The fake buffer of 100 auxiliary sync words. */
  private ByteBuffer auxSyncBuffer;

  ////////////////////////////////////////////////////////////

  public String getFormat () { return ("CMA 1A5 CHRPT"); }

  ////////////////////////////////////////////////////////////

  /** Creates a new stream using the specified file as input. */
  public CMA1A5CHRPTStream (
    String fileName
  ) throws IOException {

    // Get start year
    // --------------
    ByteBuffer input = getInputBuffer (fileName);
    input.order (ByteOrder.LITTLE_ENDIAN);
    ShortBuffer shortInput = input.asShortBuffer();
    int year = shortInput.get (FRAME_LENGTH/2 + YEAR_OFFSET);
    if (year > 70) year += 1900;
    else year += 2000;
    setYear (year);

    // Initialize stream
    // -----------------
    initStream (input, FRAME_LENGTH, FRAME_LENGTH, CHRPT_OFFSET, true);

    // Create fake buffers
    // -------------------
    spareBuffer = ByteBuffer.allocate (BLOCK_LENGTHS[SPARE_BLOCK]);

    // Create aux sync buffer
    // ----------------------
    auxSyncBuffer = ByteBuffer.allocate (BLOCK_LENGTHS[AUX_SYNC_BLOCK]);
    ShortBuffer shortBuffer = auxSyncBuffer.asShortBuffer();
    for (int i = 0; i < HRPTAuxSyncChecker.AUX_SYNC_WORDS.length; i++)
      shortBuffer.put (i, HRPTAuxSyncChecker.AUX_SYNC_WORDS[i]);

  } // CMA1A5CHRPTStream constructor

  ////////////////////////////////////////////////////////////

  protected int getBlockPosition (int block) { 

    switch (block) {
    case EARTH_BLOCK: return (EARTH_DATA_OFFSET);
    default: return (super.getBlockPosition (block));
    } // switch

  } // getBlockPosition

  ////////////////////////////////////////////////////////////

  public ByteBuffer getBlock (
    int frame,
    int block
  ) {

    switch (block) {
    case SPARE_BLOCK: return (spareBuffer.slice());
    case AUX_SYNC_BLOCK: return (auxSyncBuffer.slice());
    default: return (super.getBlock (frame, block));
    } // switch

  } // getBlock

  ////////////////////////////////////////////////////////////

  public int getDayCount (int frame) throws IOException {

    input.clear().position (FRAME_LENGTH*(frame+1) + DAY_OFFSET);
    ShortBuffer buf = input.asShortBuffer();
    int day = buf.get();
    return (day);

  } // getDayCount

  ////////////////////////////////////////////////////////////

} // CMA1A5CHRPTStream class

////////////////////////////////////////////////////////////////////////
