////////////////////////////////////////////////////////////////////////
/*
     FILE: CMA1A5HRPTStream.java
  PURPOSE: Reads CMA 1A5 format HRPT data.
   AUTHOR: Peter Hollemans
     DATE: 2007/01/27
  CHANGES: 2007/03/14, PFH, modified to extend HRPTStream

  Copyright (C) 2007, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.telemetry;

// Imports
// -------
import java.io.*;
import java.nio.*;

/**
 * The <code>CMA1A5HRPTStream</code> class reads China Meteorological
 * Administration (CMA) 1A5 format HRPT data:
 * <ul>
 *   <li>21980 byte header</li>
 *   <li>21980 bytes per HRPT minor frame composed of 1500 byte frame
 *   header and 20480 byte channel data as 10240 16-bit little endian
 *   values</li>
 * </ul><p>
 *
 * The 21980 byte header contains the following values in little endian
 * byte order:
 * <table border="1">
 *
 *   <tr>
 *     <th>Byte offset</th>
 *     <th>Java type</th>
 *     <th>Description</th>
 *   </tr>
 *
 *   <tr>
 *     <td>0</td>
 *     <td>short</td>
 *     <td>Satellite ID</td>
 *   </tr>
 *
 *   <tr>
 *     <td>2</td>
 *     <td>short</td>
 *     <td>Last two digits of starting year</td>
 *   </tr>
 *
 *   <tr>
 *     <td>4</td>
 *     <td>int</td>
 *     <td>Starting millisecond</td>
 *   </tr>
 *
 *   <tr>
 *     <td>8</td>
 *     <td>short</td>
 *     <td>Starting day</td>
 *   </tr>
 *
 *   <tr>
 *     <td>10</td>
 *     <td>short</td>
 *     <td>Last two digits of ending year</td>
 *   </tr>
 *
 *   <tr>
 *     <td>12</td>
 *     <td>int</td>
 *     <td>Ending millisecond</td>
 *   </tr>
 *
 *   <tr>
 *     <td>16</td>
 *     <td>short</td>
 *     <td>Ending day</td>
 *   </tr>
 *
 *   <tr>
 *     <td>18</td>
 *     <td>short</td>
 *     <td>Good scan line count</td>
 *   </tr>
 *
 *   <tr>
 *     <td>20</td>
 *     <td>short</td>
 *     <td>Total record count (scan line count plus header)</td>
 *   </tr>
 *
 *   <tr>
 *     <td>22</td>
 *     <td>short</td>
 *     <td>Number of errors</td>
 *   </tr>
 *
 *   <tr>
 *     <td>24</td>
 *     <td>short</td>
 *     <td>Error ratio</td>
 *   </tr>
 *
 *   <tr>
 *     <td>26</td>
 *     <td>short</td>
 *     <td>Reserved block #1</td>
 *   </tr>
 *
 *   <tr>
 *     <td>28</td>
 *     <td>short</td>
 *     <td>Time error</td>
 *   </tr>
 *
 *   <tr>
 *     <td>30</td>
 *     <td>short</td>
 *     <td>Missing scan lines</td>
 *   </tr>
 *
 *   <tr>
 *     <td>32</td>
 *     <td>short</td>
 *     <td>Ramp slope analysis result</td>
 *   </tr>
 *
 *   <tr>
 *     <td>34</td>
 *     <td>short</td>
 *     <td>Reserved block #2</td>
 *   </tr>
 *
 *   <tr>
 *     <td>36</td>
 *     <td>float[20]</td>
 *     <td>Calibration coefficients for 5 channels:
 *     <ul>
 *       <li>Slope
 *       <li>Intercept
 *       <li>Standard deviation in slope
 *       <li>Standard deviation in intercept
 *     </ul></td>
 *   </tr>
 *
 *   <tr>
 *     <td>116</td>
 *     <td>short</td>
 *     <td>Reserved block #3</td>
 *   </tr>
 *
 *   <tr>
 *     <td>118</td>
 *     <td>short</td>
 *     <td>Orbit number</td>
 *   </tr>
 *
 *   <tr>
 *     <td>120</td>
 *     <td>double[7]</td>
 *     <td>Orbital element values:
 *     <ul>
 *       <li>Epoch time
 *       <li>Orbital semi-major axis
 *       <li>Orbital ecentricity
 *       <li>Orbital inclination
 *       <li>Right ascension of ascending node
 *       <li>Argument of perigee
 *       <li>???
 *     </ul></td>
 *   </tr>
 *
 *   <tr>
 *     <td>176</td>
 *     <td>short</td>
 *     <td>Direction (0 = descending, 1 = ascending)</td>
 *   </tr>
 *
 *   <tr>
 *     <td>178</td>
 *     <td>short</td>
 *     <td>Orbit type</td>
 *   </tr>
 *
 *   <tr>
 *     <td>180</td>
 *     <td>short</td>
 *     <td>Start orbit number</td>
 *   </tr>
 *
 *   <tr>
 *     <td>182</td>
 *     <td>short[5]</td>
 *     <td>Reserved block #4</td>
 *   </tr>
 *
 *   <tr>
 *     <td>192</td>
 *     <td>double[3]</td>
 *     <td>Satellite attitude: roll, pitch, yaw</td>
 *   </tr>
 *
 *   <tr>
 *     <td>216</td>
 *     <td>short[4]</td>
 *     <td>Reserved block #5</td>
 *   </tr>
 *
 *   <tr>
 *     <td>224</td>
 *     <td>float[8]</td>
 *     <td>Latitude and longitude of 4 corner points</td>
 *   </tr>
 *
 *   <tr>
 *     <td>256</td>
 *     <td>double</td>
 *     <td>Start time in seconds since Jan 1 1980 00:00 UTC</td>
 *   </tr>
 *
 *   <tr>
 *     <td>264</td>
 *     <td>double</td>
 *     <td>End time in seconds since Jan 1 1980 00:00 UTC</td>
 *   </tr>
 *
 *   <tr>
 *     <td>272</td>
 *     <td>short</td>
 *     <td>Grid flag</td>
 *   </tr>
 *
 *   <tr>
 *     <td>274</td>
 *     <td>short</td>
 *     <td>Grid starting latitude</td>
 *   </tr>
 *
 *   <tr>
 *     <td>276</td>
 *     <td>short</td>
 *     <td>Grid starting longitude</td>
 *   </tr>
 *
 *   <tr>
 *     <td>278</td>
 *     <td>short[10851]</td>
 *     <td>Reserved block #6</td>
 *   </tr>
 *
 * </table><p>
 *
 * Each HRPT minor frame header contains the following values in little
 * endian byte order:
 * <table border="1">
 *
 *   <tr>
 *     <th>Byte offset</th>
 *     <th>Java type</th>
 *     <th>Description</th>
 *   </tr>
 *
 *   <tr>
 *     <td>0</td>
 *     <td>short</td>
 *     <td>Scan line number</td>
 *   </tr>
 *
 *   <tr>
 *     <td>2</td>
 *     <td>short</td>
 *     <td>Year</td>
 *   </tr>
 *
 *   <tr>
 *     <td>4</td>
 *     <td>int</td>
 *     <td>Millisecond</td>
 *   </tr>
 *
 *   <tr>
 *     <td>8</td>
 *     <td>short</td>
 *     <td>Day</td>
 *   </tr>
 *
 *   <tr>
 *     <td>10</td>
 *     <td>short</td>
 *     <td>Reserved block #1</td>
 *   </tr>
 *
 *   <tr>
 *     <td>12</td>
 *     <td>short</td>
 *     <td>Quality flag</td>
 *   </tr>
 *
 *   <tr>
 *     <td>14</td>
 *     <td>short</td>
 *     <td>Reserved block #2</td>
 *   </tr>
 *
 *   <tr>
 *     <td>16</td>
 *     <td>float[10]</td>
 *     <td>Slope and intercept for 5 channels</td>
 *   </tr>
 *
 *   <tr>
 *     <td>56</td>
 *     <td>float[51]</td>
 *     <td>Solar zenith angle for 51 points in degrees</td>
 *   </tr>
 *
 *   <tr>
 *     <td>260</td>
 *     <td>float[102]</td>
 *     <td>Latitude and longitude for 51 points in degrees</td>
 *   </tr>
 *
 *   <tr>
 *     <td>668</td>
 *     <td>short[103]</td>
 *     <td>First 103 words of HRPT header</td>
 *   </tr>
 *
 *   <tr>
 *     <td>874</td>
 *     <td>float[51]</td>
 *     <td>Satellite zenith angle for 51 points in degrees</td>
 *   </tr>
 *
 *   <tr>
 *     <td>1078</td>
 *     <td>float[51]</td>
 *     <td>Relative azimuth angle for 51 points in degrees</td>
 *   </tr>
 *
 *   <tr>
 *     <td>1282</td>
 *     <td>short[109]</td>
 *     <td>Reserved block #3</td>
 *   </tr>
 *
 * </table>
 */
public class CMA1A5HRPTStream 
  extends HRPTStream {

  // Constants
  // ---------

  /** Short word offset into file header for start year. */
  private static final int YEAR_OFFSET = 1;

  /** Byte offset into line header for raw HRPT data. */
  private static final int HRPT_OFFSET = 668;

  /** The minor frame length in bytes. */
  private static final int FRAME_LENGTH = 21980;

  /** The earth data offset in bytes. */
  private static final int EARTH_DATA_OFFSET = 1500 - HRPT_OFFSET;

  // Variables
  // ---------

  /** The fake buffer of 520 TIP words. */
  private ByteBuffer tipBuffer;

  /** The fake buffer of 127 spare words. */
  private ByteBuffer spareBuffer;

  /** The fake buffer of 100 auxiliary sync words. */
  private ByteBuffer auxSyncBuffer;

  ////////////////////////////////////////////////////////////

  public String getFormat () { return ("CMA 1A5 HRPT"); }

  ////////////////////////////////////////////////////////////

  /** Creates a new stream using the specified file as input. */
  public CMA1A5HRPTStream (
    String fileName
  ) throws IOException {

    // Get start year
    // --------------
    ByteBuffer input = getInputBuffer (fileName);
    input.order (ByteOrder.LITTLE_ENDIAN);
    ShortBuffer shortInput = input.asShortBuffer();
    int year = shortInput.get (YEAR_OFFSET);
    if (year > 70) year += 1900;
    else year += 2000;
    setYear (year);

    // Initialize stream
    // -----------------
    initStream (input, FRAME_LENGTH, FRAME_LENGTH, HRPT_OFFSET, true);

    // Create fake buffers
    // -------------------
    tipBuffer = ByteBuffer.allocate (BLOCK_LENGTHS[TIP_BLOCK]);
    spareBuffer = ByteBuffer.allocate (BLOCK_LENGTHS[SPARE_BLOCK]);

    // Create aux sync buffer
    // ----------------------
    auxSyncBuffer = ByteBuffer.allocate (BLOCK_LENGTHS[AUX_SYNC_BLOCK]);
    ShortBuffer shortBuffer = auxSyncBuffer.asShortBuffer();
    for (int i = 0; i < HRPTAuxSyncChecker.AUX_SYNC_WORDS.length; i++)
      shortBuffer.put (i, HRPTAuxSyncChecker.AUX_SYNC_WORDS[i]);

  } // CMA1A5HRPTStream constructor

  ////////////////////////////////////////////////////////////

  protected int getBlockPosition (int block) { 

    switch (block) {
    case EARTH_BLOCK: return (EARTH_DATA_OFFSET);
    default: return (super.getBlockPosition (block));
    } // switch

  } // getBlockPosition

  ////////////////////////////////////////////////////////////

  public ByteBuffer getBlock (
    int frame,
    int block
  ) {

    switch (block) {
    case TIP_BLOCK: return (tipBuffer.slice());
    case SPARE_BLOCK: return (spareBuffer.slice());
    case AUX_SYNC_BLOCK: return (auxSyncBuffer.slice());
    default: return (super.getBlock (frame, block));
    } // switch

  } // getBlock

  ////////////////////////////////////////////////////////////

} // CMA1A5HRPTStream class

////////////////////////////////////////////////////////////////////////
