////////////////////////////////////////////////////////////////////////
/*
     FILE: HRPTDecoder.java
  PURPOSE: Decodes NOAA HRPT data.
   AUTHOR: Peter Hollemans
     DATE: 2005/06/13
  CHANGES: 2007/01/28, PFH, added use of HRPTReaderFactory
           2007/02/20, PFH, modified to implement TelemetryDecoder
           2009/06/02, PFH, added NOAA-19 support
           2010/12/18, PFH, removed unneeded cast to int for millisecond |= ...

  Copyright (C) 2005-2010, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.telemetry;

// Imports
// -------
import java.io.*;
import java.util.*;
import java.nio.*;
import java.net.*;
import terrenus.instrument.*;
import terrenus.instrument.RadiometerCalibrator.Record;
import terrenus.orbit.*;
import terrenus.util.*;

/**
 * The <code>HRPTDecoder</code> class uses an {@link HRPTStream} to
 * decode NOAA HRPT data.
 */
public class HRPTDecoder 
  implements TelemetryDecoder {

  // Constants
  // ---------

  /** The mask for bit 1 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT1 = (short) 0x0200;

  /** The mask for bit 2 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT2 = (short) 0x0100;

  /** The mask for bit 3 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT3 = (short) 0x0080;

  /** The mask for bit 4 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT4 = (short) 0x0040;

  /** The mask for bit 5 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT5 = (short) 0x0020;

  /** The mask for bit 6 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT6 = (short) 0x0010;

  /** The mask for bit 7 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT7 = (short) 0x0008;

  /** The mask for bit 8 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT8 = (short) 0x0004;

  /** The mask for bit 9 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT9 = (short) 0x0002;

  /** The mask for bit 10 of the 10-bit field (1=MSB, 10=LSB). */
  private static final short BIT10 = (short) 0x0001;

  /** The spacecraft address for NOAA-15. */
  private static final int NOAA15_ADDRESS = 7;

  /** The spacecraft address for NOAA-16. */
  private static final int NOAA16_ADDRESS = 3;

  /** The spacecraft address for NOAA-17. */
  private static final int NOAA17_ADDRESS = 11;

  /** The spacecraft address for NOAA-18. */
  private static final int NOAA18_ADDRESS = 13;

  /** The spacecraft address for NOAA-19. */
  private static final int NOAA19_ADDRESS = 15;

  /** The number of milliseconds between adjacent frames. */
  private static final double MSEC_PER_FRAME = 1000.0/6;

  /** The number of milliseconds in one hour. */
  private static final int MSEC_PER_HOUR = 3600000;

  /** The number of milliseconds in one minute. */
  private static final int MSEC_PER_MINUTE = 60000;

  /** The number of milliseconds in one second. */
  private static final int MSEC_PER_SECOND = 1000;

  // Variables
  // ---------

  /** The stream for obtaining data. */
  private HRPTStream stream;

  /** The actual spacecraft address. */
  private int spacecraftId;

  /** The spacecraft corresponding to the spacecraft address. */
  private Spacecraft spacecraft;

  /** The actual data start time in milliseconds since the epoch. */
  private long startTime;

  /** The array of frame sync errors. */
  private int[] frameSyncErrors;

  /** The array of auxiliary sync errors. */
  private int[] auxSyncErrors;

  /** The array of TIP parity errors. */
  private int[] tipParityErrors;

  /** The calibrator for AVHRR calibration. */
  private AVHRRCalibrator avhrrCalibrator;

  /** The locator for AVHRR location. */
  private AVHRRLocator avhrrLocator;

  ////////////////////////////////////////////////////////////

  static {

    /**
     * To aid in using NOAA HRPT data, we add various sources of
     * orbital elements here.  That way, other classes can make use of
     * them through the {@link OrbitalElementManager} without having
     * to know where they came from.
     */
    try {

      // Add archive sources
      // -------------------
      OrbitalElementManager manager = OrbitalElementManager.getInstance();
      manager.registerSource (new ArchiveElementSource (
        OrbitalElementManager.class.getResource ("noaa15.txt")));
      manager.registerSource (new ArchiveElementSource (
        OrbitalElementManager.class.getResource ("noaa16.txt")));
      manager.registerSource (new ArchiveElementSource (
        OrbitalElementManager.class.getResource ("noaa17.txt")));
      manager.registerSource (new ArchiveElementSource (
        OrbitalElementManager.class.getResource ("noaa18.txt")));
      manager.registerSource (new ArchiveElementSource (
        OrbitalElementManager.class.getResource ("noaa19.txt")));

      // Add current sources
      // -------------------
      manager.registerSource (new CurrentElementSource (
        new URL ("http://www.celestrak.net/NORAD/elements/noaa.txt")));

    } // try
    catch (IOException e) {
      throw new RuntimeException ("Error loading default orbital elements");
    } // catch

  } // static

  ////////////////////////////////////////////////////////////

  public boolean hasErrors (int frame) throws IOException {

    int minorFrame = getMinorFrame (frame);
    boolean errors = (
      frameSyncErrors[frame] != 0 ||
      auxSyncErrors[frame] != 0 ||
      (getMinorFrame (frame) != 2 && tipParityErrors[frame] != 0)
    );
    return (errors);

  } // hasErrors

  ////////////////////////////////////////////////////////////

  /** Create a new decoder using the specified stream. */
  public HRPTDecoder (HRPTStream stream) throws IOException { 

    this.stream = stream; 

    // Create frame error arrays
    // -------------------------
    int frameCount = stream.getFrameCount();
    frameSyncErrors = new int[frameCount];
    auxSyncErrors = new int[frameCount];
    tipParityErrors = new int[frameCount];    

    // Initialize statistical tests
    // ----------------------------
    FrequencyMap<Integer> spacecraftMap = new FrequencyMap<Integer>();
    ArrayList<Long> startTimeList = new ArrayList<Long>();
    Calendar cal = new GregorianCalendar (TimeZone.getTimeZone ("GMT+0"));
    int year = stream.getYear();
    if (year == -1) 
      throw new RuntimeException ("Cannot determine year from HRPT data");

    // Loop over each frame
    // --------------------
    for (int i = 0; i < frameCount; i++) {

      // Compute bit errors
      // ------------------
      frameSyncErrors[i] = computeFrameSyncErrors (i);
      auxSyncErrors[i] = computeAuxSyncErrors (i);
      tipParityErrors[i] = computeTIPParityErrors (i);

      // Check if data is usable
      // -----------------------
      boolean isUsable = !hasErrors (i) && isAVHRRSync (i) && !isResync (i);
      if (!isUsable) continue;

      // Add spacecraft to frequency map
      // -------------------------------
      spacecraftMap.put (getSpacecraftId (i));

      // Add start time to list
      // ----------------------
      cal.set (Calendar.YEAR, year);
      cal.set (Calendar.DAY_OF_YEAR, getDayCount (i));
      int millisecond = getMillisecond (i);
      int hour = millisecond/MSEC_PER_HOUR;
      cal.set (Calendar.HOUR_OF_DAY, hour);
      int minute = (millisecond - hour*MSEC_PER_HOUR)/MSEC_PER_MINUTE;
      cal.set (Calendar.MINUTE, minute);
      int second = (millisecond - hour*MSEC_PER_HOUR - 
        minute*MSEC_PER_MINUTE)/MSEC_PER_SECOND;
      cal.set (Calendar.SECOND, second);
      int newMilli = millisecond - hour*MSEC_PER_HOUR - minute*MSEC_PER_MINUTE 
        - second*MSEC_PER_SECOND;
      cal.set (Calendar.MILLISECOND, newMilli);
      long thisStartTime = (long) (cal.getTimeInMillis() - i*MSEC_PER_FRAME);
      startTimeList.add (thisStartTime);

    } // for

    // Set spacecraft ID
    // -----------------
    Integer freqValue = spacecraftMap.getFreqValue();
    if (freqValue == null)
      throw new IOException ("Cannot locate valid HRPT data");
    spacecraftId = freqValue;

    // Set start time
    // --------------
    Collections.sort (startTimeList);
    int median = startTimeList.size()/2;
    startTime = startTimeList.get (median);

  } // HRPTDecoder

  ////////////////////////////////////////////////////////////

  public HRPTStream getStream () { return (stream); }

  ////////////////////////////////////////////////////////////

  /** Computes the number of errors in the frame sync bits. */
  private int computeFrameSyncErrors (int frame) throws IOException {

    ByteBuffer buf = stream.getBlock (frame, HRPTStream.FRAME_SYNC_BLOCK);
    return (HRPTFrameSyncChecker.getInstance().getDifferenceCount (buf));
    
  } // computeFrameSyncErrors

  ////////////////////////////////////////////////////////////

  /**
   * Gets the number of errors in the frame sync bits, or 0 if no
   * errors are present.  The number of errors is the total number of
   * bits which differ between the expected frame sync bits and the
   * actual frame sync bits.
   */
  public int getFrameSyncErrors (int frame) { 

    return (frameSyncErrors[frame]); 

  } // getFrameSyncErrors

  ////////////////////////////////////////////////////////////

  /** 
   * Returns true if the frame is an AVHRR sync, or false if it is an
   * internal sync.
   */
  public boolean isAVHRRSync (int frame) throws IOException { 

    ShortBuffer buf = 
      stream.getBlock (frame, HRPTStream.ID_BLOCK).asShortBuffer();
    return ((buf.get(0) & BIT1) != 0); 

  } // isAVHRRSync

  ////////////////////////////////////////////////////////////

  /**
   * Returns the minor frame number, either 1, 2, or 3. 
   */
  public int getMinorFrame (int frame) throws IOException { 
    
    ShortBuffer buf = 
      stream.getBlock (frame, HRPTStream.ID_BLOCK).asShortBuffer();
    return ((buf.get(0) & (BIT2|BIT3)) >>> 7);

  } // getMinorFrame
            
  ////////////////////////////////////////////////////////////

  /**
   * Gets the start time of frame capture in milliseconds since the
   * epoch.  This time may differ from the values returned by {@link
   * #getDayCount} and {@link #getMillisecond} since some frames may
   * have bit errors.  The actual frame time is computed by examining
   * a number of frame times, inferring the correct data start time
   * from those times, and then computing the frame time using the
   * frame index and the known frame step interval.  This presents the
   * HRPT data as a continuous, uniformly stepping set of frames and
   * is very useful for computing correct Earth locations for every
   * sensor scan even when the frame itself contains unusable sensor
   * data.
   *
   * @param frame the frame to return the time for, starting at 0.
   *
   * @return the frame start time in milliseconds since the epoch.
   * The epoch in this case is January 1, 1970 00:00:00 UTC.
   */
  public long getTime (int frame) {

    return ((long) (startTime + frame*MSEC_PER_FRAME));

  } // getTime

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the spacecraft address identified for this HRPT data.
   * This may differ from the spacecraft address returned by
   * {@link #getSpacecraftId(int)} since some frames may have bit
   * errors.  The actual spacecraft address is detected by
   * examining a number of HRPT minor frames.
   *
   * @return the actual spacecraft address for this data.
   */
  public int getSpacecraftId () { return (spacecraftId); }

  ////////////////////////////////////////////////////////////

  /** Gets the spacecraft address for the frame. */
  public int getSpacecraftId (int frame) throws IOException {

    ShortBuffer buf = 
      stream.getBlock (frame, HRPTStream.ID_BLOCK).asShortBuffer();
    return ((buf.get(0) & (BIT4|BIT5|BIT6|BIT7)) >>> 3);

  } // getSpacecraftId

  ////////////////////////////////////////////////////////////

  /** 
   * Returns true if the frame is a resync, or false if it is
   * stable.
   */
  public boolean isResync (int frame) throws IOException { 

    ShortBuffer buf = 
      stream.getBlock (frame, HRPTStream.ID_BLOCK).asShortBuffer();
    return ((buf.get(0) & BIT8) != 0); 

  } // isResync

  ////////////////////////////////////////////////////////////

  /** 
   * Returns true if the frame is normal AVHRR input, or false
   * if it is pseudonoise.
   */
  public boolean isAVHRRNormal (int frame) throws IOException { 

    ShortBuffer buf = 
      stream.getBlock (frame, HRPTStream.ID_BLOCK).asShortBuffer();
    return ((buf.get(0) & BIT9) != 0); 

  } // isAVHRRNormal

  ////////////////////////////////////////////////////////////

  /** 
   * Returns true if the frame contains AVHRR channel 3a data, or
   * false if it contains AVHRR channel 3b.
   */
  public boolean isChannel3a (int frame) throws IOException { 

    ShortBuffer buf = 
      stream.getBlock (frame, HRPTStream.ID_BLOCK).asShortBuffer();
    return ((buf.get(0) & BIT10) != 0); 

  } // isChannel3a

  ////////////////////////////////////////////////////////////

  /** Gets the frame day count. */
  public int getDayCount (int frame) throws IOException {

    ShortBuffer buf = 
      stream.getBlock (frame, HRPTStream.TIME_CODE_BLOCK).asShortBuffer();
    return (buf.get(0) >>> 1);

  } // getDayCount

  ////////////////////////////////////////////////////////////

  /** Gets the frame millisecond. */
  public int getMillisecond (int frame) throws IOException {

    ShortBuffer buf = 
      stream.getBlock (frame, HRPTStream.TIME_CODE_BLOCK).asShortBuffer();
    int millisecond = 0;
    millisecond |= (buf.get(1) & 0x007f) << 20;
    millisecond |= ((int) buf.get(2)) << 10;
    millisecond |= buf.get(3);
    return (millisecond);

  } // getMillisecond

  ////////////////////////////////////////////////////////////

  /** 
   * Returns true if the frame AVHRR sync delta is late, or
   * false if it is early.
   */
  public boolean isAVHRRSyncLate (int frame) throws IOException { 

    ShortBuffer buf = 
      stream.getBlock (frame, HRPTStream.SYNC_DELTA_BLOCK).asShortBuffer();
    return ((buf.get(0) & BIT1) != 0); 

  } // isAVHRRSyncLate

  ////////////////////////////////////////////////////////////

  /** Gets the count of 0.9984 MHz AVHRR sync periods. */
  public int getAVHRRSyncPeriods (int frame) throws IOException { 

    ShortBuffer buf = 
      stream.getBlock (frame, HRPTStream.SYNC_DELTA_BLOCK).asShortBuffer();
    return (buf.get(0) & (~BIT1));

  } // getAVHRRSyncPeriods

  ////////////////////////////////////////////////////////////

  /** 
   * Computes the number of TIP words in the frame that have
   * errors based on an even parity check over bits 1-8 with parity in
   * bit 9.
   */
  private int computeTIPParityErrors (int frame) throws IOException {

    ShortBuffer buf = 
      stream.getBlock (frame, HRPTStream.TIP_BLOCK).asShortBuffer();
    int errors = 0;
    while (buf.hasRemaining()) {
      short word = (short) (buf.get() >>> 1);
      boolean isEven = true;
      for (int bit = 0; bit < 9; bit++) {
        if (((0x1 << bit) & word) != 0) isEven = !isEven;
      } // for
      if (!isEven) errors++;
    } // for
    return (errors);
    
  } // computeTIPParityErrors

  ////////////////////////////////////////////////////////////

  /**
   * Gets the number of TIP words in the frame that have
   * errors based on an even parity check over bits 1-8 with parity in
   * bit 9.
   */
  public int getTIPParityErrors (int frame) {

    return (tipParityErrors[frame]);

  } // getTIPParityErrors

  ////////////////////////////////////////////////////////////

  /** Computes the number of errors in the frame auxiliary sync bits. */
  private int computeAuxSyncErrors (int frame) throws IOException {

    ByteBuffer buf = stream.getBlock (frame, HRPTStream.AUX_SYNC_BLOCK);
    return (HRPTAuxSyncChecker.getInstance().getDifferenceCount (buf));
    
  } // computeAuxSyncErrors

  ////////////////////////////////////////////////////////////

  /**
   * Gets the number of errors in the frame auxiliary sync bits, or
   * 0 if no errors are present.  The number of errors is the total
   * number of bits which differ between the expected auxiliary sync
   * bits and the actual auxiliary sync bits.
   */
  public int getAuxSyncErrors (int frame) {

    return (auxSyncErrors[frame]);

  } // getAuxSyncErrors

  ////////////////////////////////////////////////////////////

  public InstrumentData getData (
    int frame,
    Instrument inst
  ) throws IOException {

    // Check instrument
    // ----------------
    if (!(inst instanceof AVHRR)) 
      throw new IllegalArgumentException ("Illegal instrument: " + 
        inst.getClass().getName());

    // Create AVHRR objects
    // --------------------
    if (avhrrCalibrator == null) avhrrCalibrator = new AVHRRCalibrator (this);
    if (avhrrLocator == null) avhrrLocator = new AVHRRLocator (this);

    // Create AVHRR data for frame
    // ---------------------------
    Record record = avhrrCalibrator.getCalibration (frame);
    AVHRRData data = new AVHRRData (frame, this, avhrrCalibrator, 
      record, avhrrLocator);

    return (data);

  } // getData

  ////////////////////////////////////////////////////////////

  public Spacecraft getSpacecraft () { 

    // Create new spacecraft object
    // ----------------------------
    if (spacecraft == null) {
      int noradId;
      switch (spacecraftId) {
      case NOAA15_ADDRESS: noradId = 25338; break;
      case NOAA16_ADDRESS: noradId = 26536; break;
      case NOAA17_ADDRESS: noradId = 27453; break;
      case NOAA18_ADDRESS: noradId = 28654; break;
      case NOAA19_ADDRESS: noradId = 33591; break;
      default: throw new RuntimeException ("Unknown spacecraft address: " + 
        spacecraftId);
      } // switch
      spacecraft = new NOAAKLMSpacecraft (noradId);
    } // if

    return (spacecraft);

  } // getSpacecraft

    ////////////////////////////////////////////////////////////
    
    /** Tests this class. */
    public static void main (String[] argv) throws Exception {
      
      // Create stream
      // -------------
      String name = argv[0];
      TelemetryStream stream = TelemetryStreamFactory.create (name);
      if (argv.length > 1)
        stream.setYear (Integer.parseInt (argv[1]));
      
      // Create calibration records
      // --------------------------
      HRPTDecoder decoder = (HRPTDecoder) stream.getDecoder();
      int lines = stream.getFrameCount();
      for (int i = 0; i < lines; i++) {
        if (i % 40 == 0) {
          for (int j = 0; j < 9*11; j++) System.out.print ("-");
          System.out.println();
          System.out.printf (
                           "%8s | %8s | %8s | %8s | %8s | %8s | %8s | %8s | %8s\n",
                           "Line",
                           "isUsable",
                           "hasErr",
                           "fSyncErr",
                           "aSyncErr",
                           "tParErr",
                           "isAVSync",
                           "isResync",
                           "isAVNorm"
                           );
          for (int j = 0; j < 9*11; j++) System.out.print ("-");
          System.out.println();
        } // if
        boolean hasErrors = decoder.hasErrors (i);
        boolean isAVHRRSync = decoder.isAVHRRSync (i);
        boolean isResync = decoder.isResync (i);
        boolean isUsable = !hasErrors && isAVHRRSync && !isResync;
        int frameSyncErrors = decoder.getFrameSyncErrors (i);
        int auxSyncErrors = decoder.getAuxSyncErrors (i);
        int tipParityErrors = decoder.getTIPParityErrors (i);
        if (decoder.getMinorFrame (i) == 2) tipParityErrors = 0;
        boolean isAVHRRNormal = decoder.isAVHRRNormal (i);        
        System.out.printf (
                            "%8d | %8s | %8s | %8d | %8d | %8d | %8s | %8s | %8s\n",
                            i,
                            (isUsable ? "YES" : "NO"),
                            (hasErrors ? "YES" : "NO"),
                            frameSyncErrors,
                            auxSyncErrors,
                            tipParityErrors,
                            (isAVHRRSync ? "YES" : "NO"),
                            (isResync ? "YES" : "NO"),
                            (isAVHRRNormal ? "YES" : "NO")
                            );
      } // for
      
    } // main
    
    ////////////////////////////////////////////////////////////

} // HRPTDecoder class

////////////////////////////////////////////////////////////////////////
