////////////////////////////////////////////////////////////////////////
/*
     FILE: HRPTStream.java
  PURPOSE: Interface for reading HRPT data.
   AUTHOR: Peter Hollemans
     DATE: 2005/06/13
  CHANGES: 2007/02/09, PFH, modifid to extend TelemetryStream

  Copyright (C) 2005, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.telemetry;

// Imports
// -------
import java.io.*;
import java.nio.*;

/**
 * The <code>HRPTStream</code> is a telemetry stream of NOAA HRPT
 * data.  The stream delivers HRPT data as detailed in "NOAA Technical
 * Memorandum NESS 107 - Rev. 1, Data Extraction and Calibration of
 * TIROS-N/NOAA Radiometers, Walter G. Planet, Washington, D.C.,
 * November 1979 (Revised October 1988)".<p>
 *
 * The methods commit the subclass to the delivery of HRPT minor
 * frames in the same format as outlined in the NOAA documentation,
 * regardless of the underlying capture station storage format.  In
 * the case where the storage format does not store some part of the
 * data, the methods may create artificial data (such as the frame
 * synchronization) or may resort to zero-filling if there is no
 * reasonable alternative.  The only modification to the original
 * documented format is that words returned by the methods contain
 * data in the least significant 10 bits of a 16-bit value rather than
 * the original 10-bit word.<p>
 *
 * Subclasses must implement the {@link #getFormat} method and
 * override any other methods such as {@link #getFramePosition}
 * that would required changes if the file format is structurally
 * very different from a broadcast HRPT stream.
 */
public abstract class HRPTStream
  extends AbstractTelemetryStream {

  // Constants
  // ---------

  /** The array of HRPT frame block offsets in bytes. */
  public static final int[] BLOCK_OFFSETS = new int[] {
    0*2,     // Frame sync
    6*2,     // ID (AVHRR)
    8*2,     // Time code
    12*2,    // Telemetry (AVHRR)
    22*2,    // Internal target data (AVHRR)
    52*2,    // Space data (AVHRR)
    102*2,   // Sync delta (AVHRR)
    103*2,   // TIP data
    623*2,   // Spare words
    750*2,   // Earth data (AVHRR)
    10990*2  // Auxiliary sync
  };

  /** The array of HRPT frame block lengths in bytes. */
  public static final int[] BLOCK_LENGTHS = new int[] {
    6*2,     // Frame sync
    2*2,     // ID (AVHRR)
    4*2,     // Time code
    10*2,    // Telemetry (AVHRR)
    30*2,    // Internal target data (AVHRR)
    50*2,    // Space data (AVHRR)
    1*2,     // Sync delta (AVHRR)
    520*2,   // TIP data
    127*2,   // Spare words
    10240*2, // Earth data (AVHRR)
    100*2    // Auxiliary sync
  };

  /** Frame sync block index. */
  public static final int FRAME_SYNC_BLOCK = 0;

  /** ID block index. */
  public static final int ID_BLOCK = 1;

  /** Time code block index. */
  public static final int TIME_CODE_BLOCK = 2;

  /** Telemetry block index. */
  public static final int TELEMETRY_BLOCK = 3;

  /** Internal target block index. */
  public static final int INTERNAL_TARGET_BLOCK = 4;

  /** Space data block index. */
  public static final int SPACE_BLOCK = 5;

  /** Sync delta block index. */
  public static final int SYNC_DELTA_BLOCK = 6;

  /** TIP data block index. */
  public static final int TIP_BLOCK = 7;

  /** Spare words block index. */
  public static final int SPARE_BLOCK = 8;

  /** Earth data block index. */                                              
  public static final int EARTH_BLOCK = 9;

  /** Auxiliary sync block index. */
  public static final int AUX_SYNC_BLOCK = 10;

  ////////////////////////////////////////////////////////////

  public int getBlockCount() { return (BLOCK_OFFSETS.length); }

  ////////////////////////////////////////////////////////////

  protected int getBlockPosition (int block) { return (BLOCK_OFFSETS[block]); }

  ////////////////////////////////////////////////////////////

  protected int getBlockLength (int block) { return (BLOCK_LENGTHS[block]); }

  ////////////////////////////////////////////////////////////

  public TelemetryDecoder getDecoder() throws IOException { 

    return (new HRPTDecoder (this)); 

  } // getDecoder

  ////////////////////////////////////////////////////////////

  protected boolean checkFrame (
    int frame
  ) {

    ByteBuffer buffer = getBlock (frame, FRAME_SYNC_BLOCK);
    int diff = HRPTFrameSyncChecker.getInstance().getDifferenceCount (buffer);
    if (diff == 0) return (true);
    else return (false);

  } // checkFrame

  ////////////////////////////////////////////////////////////

} // HRPTStream class

////////////////////////////////////////////////////////////////////////
