////////////////////////////////////////////////////////////////////////
/*
     FILE: NOAA1bHRPTStream.java
  PURPOSE: Reads NOAA 1b data format HRPT data.
   AUTHOR: Peter Hollemans
     DATE: 2010/08/28
  CHANGES: n/a

  Copyright (C) 2010, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.telemetry;

// Imports
// -------
import java.io.*;
import java.nio.*;

/**
 * The <code>NOAA1bHRPTStream</code> class reads NOAA 1b format data
 * files and extracts just the HRPT frame data.  The format supported
 * has two variants:
 * <ol>

 *   <li>NOAA 1b version 1 style (pre-KLM) as described at:<br>
 *   http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/podug/html/c2/sec2-0.htm<br>
 *   http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/podug/html/c2/sec2-1.htm<br>
 *   http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/podug/html/c3/sec3-1.htm#sec3-121<br>
 *   http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/podug/html/c3/sec3-2.htm<br>
 *   <ul>
 *     <li>122 byte Terabit Memory (TBM) header (format in Section 2.1.1)</li>
 *     <li>14800 byte dataset header (format in Section 2.0.4)</li>
 *     <li>14800 byte dataset records (format in Section 3.2.2.1)</li>
 *     <li>140 bytes of HRPT telemetry starting at byte offset 308 within each 
 *     dataset record (format in Section 3.1.2.1).  The initial 103 10-bit 
 *     HRPT telemetry words are stored packed as three 10-bit words in four 
 *     bytes, right justified (the last four byte group only contains one 10-bit
 *     word).</li>
 *     <li>13656 bytes of HRPT earth data starting at byte offset 448 within 
 *     each data record.  The 10240 10-bit HRPT earth scan words are stored
 *     packed, analagous to the initial 103 telemetry words.</li>
 *   </ul>
 *   </li>
 *
 *   <li>NOAA 1b version 2 and up style (KLMNN') as described at:<br>
 *   http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/klm/html/c8/sec831-1.htm<br>
 *   http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/klm/html/c8/sec831-2.htm<br>
 *   http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/klm/html/c8/sec83132-2.htm<br>
 *   http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/klm/html/c8/sec83133-2.htm<br>
 *   <ul>
 *     <li>512 byte Archive Retrieval System (ARS) header (for in Section 8.3.1.2)</li>
 *     <li>15872 byte dataset header (format in Section 8.3.1.3.2.2)</li>
 *     <li>15872 byte dataset records (format in Section 8.3.1.3.3.2)</li>
 *     <li>206 bytes of HRPT telemetry starting at byte offset 1056 within each 
 *     dataset record.  The initial 103 10-bit HRPT telemetry words are stored 
 *     as one 10-bit word in two bytes, right justified.</li>
 *     <li>13656 bytes of HRPT earth data starting at byte offset 1264 within 
 *     each data record.  The 10240 10-bit HRPT earth scan words are stored
 *     packed, analagous to the initial 103 telemetry words in NOAA 1b version 1 
 *     style.</li>
 *   </ul>
 *   </li>
 *
 *   </ol>
 *   <p>
 */
public class NOAA1bHRPTStream 
  extends HRPTStream {

  // Constants
  // ---------

  /** Short word offset into file header for start year. */
  private static final int YEAR_OFFSET = 1;

  /** The file header length in bytes. */
  private static final int FILE_HEADER_LENGTH_V1 = 122;
  private static final int FILE_HEADER_LENGTH_V2 = 512;
  
  /** Byte offset into dataset record of initial HRPT data. */
  private static final int HRPT_OFFSET_V1 = 308;
  private static final int HRPT_OFFSET_V2 = 1056;

  /** Byte offset into dataset record for earth scan HRPT data. */
  private static final int EARTH_DATA_OFFSET_V1 = 448;
  private static final int EARTH_DATA_OFFSET_V2 = 1264;

  /** The data record length in bytes. */
  private static final int RECORD_LENGTH_V1 = 14800;
  private static final int RECORD_LENGTH_V2 = 15872;

  


  // Variables
  // ---------

  /** The fake buffer of 520 TIP words. */
  private ByteBuffer tipBuffer;

  /** The fake buffer of 127 spare words. */
  private ByteBuffer spareBuffer;

  /** The fake buffer of 100 auxiliary sync words. */
  private ByteBuffer auxSyncBuffer;

  /** The format variant, either 1 or 2. */
  private int formatVariant;
  
  


  ////////////////////////////////////////////////////////////

  public String getFormat () { return ("NOAA 1b HRPT"); }

  ////////////////////////////////////////////////////////////

  /**
   * Gets the format variant.
   *
   * @param input the input byte buffer.
   *
   * @return the format variant, either 1, 2, or -1 for
   * unknown variant.
   */
  private int getFormatVariant (
    ByteBuffer input
  ) {
  
  
  
    return (1);
  
  } // getFormatVariant

  ////////////////////////////////////////////////////////////

  /** Creates a new stream using the specified file as input. */
  public NOAA1bHRPTStream (
    String fileName
  ) throws IOException {

    // Get start year
    // --------------
/*
    ByteBuffer input = getInputBuffer (fileName);
    input.order (ByteOrder.LITTLE_ENDIAN);
    ShortBuffer shortInput = input.asShortBuffer();
    int year = shortInput.get (YEAR_OFFSET);
    if (year > 70) year += 1900;
    else year += 2000;
    setYear (year);
*/

    // Determine format variant
    // ------------------------
    int headerLength, frameLength, hrptOffset;
    ByteBuffer input = getInputBuffer (fileName);
    formatVariant = getFormatVariant (input);
    if (formatVariant == 1) {
      headerLength = FILE_HEADER_LENGTH_V1 + RECORD_LENGTH_V1;
      frameLength = RECORD_LENGTH_V1;
      hrptOffset = HRPT_OFFSET_V1;
    } // if
    else if (formatVariant == 2) {
      headerLength = FILE_HEADER_LENGTH_V2 + RECORD_LENGTH_V2;
      frameLength = RECORD_LENGTH_V2;
      hrptOffset = HRPT_OFFSET_V2;
    } // else if
    else {
      throw new IOException ("Unknown NOAA 1b format variant");
    } // else

    // Initialize stream
    // -----------------
    initStream (input, headerLength, frameLength, hrptOffset, false);

    // Create fake buffers
    // -------------------
    tipBuffer = ByteBuffer.allocate (BLOCK_LENGTHS[TIP_BLOCK]);
    spareBuffer = ByteBuffer.allocate (BLOCK_LENGTHS[SPARE_BLOCK]);

    // Create aux sync buffer
    // ----------------------
    auxSyncBuffer = ByteBuffer.allocate (BLOCK_LENGTHS[AUX_SYNC_BLOCK]);
    ShortBuffer shortBuffer = auxSyncBuffer.asShortBuffer();
    for (int i = 0; i < HRPTAuxSyncChecker.AUX_SYNC_WORDS.length; i++)
      shortBuffer.put (i, HRPTAuxSyncChecker.AUX_SYNC_WORDS[i]);

  } // NOAA1bHRPTStream constructor

  ////////////////////////////////////////////////////////////

  /**
   * Unpacks a buffer of 10-bit words into one with 16-bit words.  The 10-bit
   * words are packed three words to fours bytes in the least significant
   * 30 bits.
   *
   * @param packedBuffer the packed buffer with three 10-bit words per four 
   * bytes.
   *
   * @return the unpacked buffer with one 10-bit word per two bytes.
   */
  private ByteBuffer getUnpacked (
    ByteBuffer packedBuffer
  ) {
  
    // Create int and short buffers
    // ----------------------------
    IntBuffer intBuffer = packedBuffer.asIntBuffer();
    ByteBuffer unpackedBuffer = ByteBuffer.allocate (intBuffer.capacity()*3);
    ShortBuffer shortBuffer = unpackedBuffer.asShortBuffer();

    // Unpack data values
    // ------------------
    while (intBuffer.hasRemaining()) {
      int word = intBuffer.get();
      shortBuffer.put ((short) ((0x3ff00000 & word) >>> 20));
      shortBuffer.put ((short) ((0x000ffc00 & word) >>> 10));
      shortBuffer.put ((short) (0x000003ff & word));
    } // while
    
    return (unpackedBuffer);

  } // getUnpacked

  ////////////////////////////////////////////////////////////

  public ByteBuffer getBlock (
    int frame,
    int block
  ) {

    switch (block) {

    // Handle missing data in either variant
    // -------------------------------------
    case TIP_BLOCK: return (tipBuffer.slice());
    case SPARE_BLOCK: return (spareBuffer.slice());
    case AUX_SYNC_BLOCK: return (auxSyncBuffer.slice());

    default: 

      // Handle format version 1 unpacking
      // ---------------------------------
      if (formatVariant == 1) {
        ByteBuffer blockBuffer = ByteBuffer.allocate (BLOCK_LENGTHS[block]);

      
      
      } // if
      
      // Handle format version 2 unpacking
      // ---------------------------------
      else {
      
      
      
      } // else
    
//    return (super.getBlock (frame, block));
    } // switch

    return (null);


  } // getBlock

  ////////////////////////////////////////////////////////////

} // NOAA1bHRPTStream class

////////////////////////////////////////////////////////////////////////
