////////////////////////////////////////////////////////////////////////
/*
     FILE: QuorumFile.java
  PURPOSE: Reads Quorum data files.
   AUTHOR: Peter Hollemans
     DATE: 2007/02/11
  CHANGES: 2010/12/18, PFH, removed unneeded cast to (String[]) in ephemeris.clone()

  Copyright (C) 2007-2010, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.telemetry;

// Imports
// -------
import java.io.*;
import java.nio.*;
import java.nio.channels.*;
import java.util.*;
import java.text.*;
import terrenus.orbit.*;

/**
 * The <code>QuorumFile</code> class reads Quorum QTrack data files
 * and reports various information such as the ephermeris data and
 * pseudoframe positions.
 */
public class QuorumFile
  implements OrbitalElementSource {

  // Constants
  // ---------

  /** The pseudo frame marker. */
  private static final short PSEUDO = (short) 0x8000;

  /** The pseudo frame ephemeris type marker. */
  private static final short PSEUDO_EPHEMERIS = (short) 0x8000;
  
  /** The pseudo frame start of data type marker. */
  private static final short PSEUDO_START = (short) 0x8100;
  
  /** The pseudo frame resync type marker. */
  private static final short PSEUDO_RESYNC = (short) 0x8200;
  
  /** The pseudo frame file close type marker. */
  private static final short PSEUDO_CLOSE = (short) 0x8300;

  // Variables
  // ---------

  /** The position in words of each minor frame. */
  private int[] framePosition;

  /** The ephemeris strings. */
  private String[] ephemeris;

  /** The data start time. */
  private long startTime;
  
  ////////////////////////////////////////////////////////////

  /** 
   * Creates a new Quorum file.
   *
   * @param fileName the name of the data file.
   * @param frameWords the number of 16-bit words in each frame.
   */
  public QuorumFile (
    String fileName,
    int frameWords
  ) throws IOException {

    // Get file channel
    // ----------------
    File file = new File (fileName);
    long fileLength = file.length();
    RandomAccessFile raFile = new RandomAccessFile (file, "r");
    FileChannel channel = raFile.getChannel();

    // Create a buffer of the starting data frames
    // -------------------------------------------
    int frameBytes = frameWords*2;
    int maxFrames = (int) Math.min (channel.size()/frameBytes, 100);
    ShortBuffer startBuf = channel.map (FileChannel.MapMode.READ_ONLY,
      0, maxFrames*frameBytes).order (ByteOrder.LITTLE_ENDIAN).asShortBuffer();

    // Search for the actual data start
    // --------------------------------
    int startPosition = -1;
    for (int i = 0; i < startBuf.capacity()-1; i++) {
      if (startBuf.get (i) == PSEUDO && startBuf.get(i+1) == PSEUDO_EPHEMERIS)
        startPosition = i;
    } // for
    if (startPosition == -1)
      throw new IOException ("Unable to find ephemeris pseudoframe");

    // Read ephemeris strings
    // ----------------------
    ephemeris = new String[2];
    startBuf.position (startPosition + 3);
    for (int line = 0; line < 3; line++) {
      StringBuffer buffer = new StringBuffer();
      char ch;
      while ((ch = (char)(startBuf.get() & 0xff)) != 0)
        buffer.append (ch);
      if (line > 0) ephemeris[line-1] = buffer.toString();
    } // for

    // Find pseudo frames
    // ------------------
    List<Integer> pseudoPositionList = new ArrayList<Integer>();
    int frameCount = ((int) (channel.size()/2) - startPosition)/frameWords;
    ByteBuffer byteBuf = ByteBuffer.allocate (4).order (
      ByteOrder.LITTLE_ENDIAN);
    ShortBuffer shortBuf = byteBuf.asShortBuffer();
    for (int i = 0; i < frameCount; i++) {

      // Read first two words of frame
      // -----------------------------
      byteBuf.clear();
      channel.read (byteBuf, (startPosition + i*frameWords)*2);

      // Check for pseudoframe
      // ---------------------
      short word1 = shortBuf.get (0);
      if (word1 == PSEUDO)
        pseudoPositionList.add (new Integer (i));

      // Check for start pseudoframe
      // ---------------------------
      short word2 = shortBuf.get (1);
      if (word2 == PSEUDO_START) {
        byteBuf.clear();
        channel.read (byteBuf, (startPosition + i*frameWords + 3)*2);
        short timeWord1 = shortBuf.get (0);
        short timeWord2 = shortBuf.get (1);
        long time = (((long) (timeWord1 & 0xffff)) << 16) | 
          ((long) (timeWord2 & 0xffff));
        startTime = time*1000;
      } // if

    } // while

    // Close the file
    // --------------
    raFile.close();

    // Create normal frame position array
    // ----------------------------------
    framePosition = new int[frameCount - pseudoPositionList.size()];
    Iterator<Integer> pseudoIter = pseudoPositionList.iterator();
    int nextPseudo = (pseudoIter.hasNext() ? pseudoIter.next() : -1);
    int pseudoCount = 0;
    for (int i = 0; i < frameCount; i++) {
      while (nextPseudo == i) {
        pseudoCount++;
        nextPseudo = (pseudoIter.hasNext() ? pseudoIter.next() : -1);
        i++;
      } // while
      if (i < frameCount)  
        framePosition[i-pseudoCount] = startPosition + i*frameWords;
    } // for

  } // QuorumFile constructor

  ////////////////////////////////////////////////////////////

  public String[] getElements (
    int satid,
    Date date,
    double maxDays
  ) {

    // Check satellite id
    // ------------------
    int id = -1;
    try { id = Integer.parseInt (ephemeris[0].substring (2, 7)); }
    catch (NumberFormatException e) { }
    if (id != satid) return (null);

    // Check date
    // ----------
    double days = AbstractOrbitalElementSource.getEpochOffset (ephemeris, 
      date);
    if (Math.abs (days) > maxDays) return (null);

    return (ephemeris.clone());

  } // getElements

  ////////////////////////////////////////////////////////////

  /**
   * Gets the ephemeris strings.
   *
   * @return the aray of two ephemeris strings.
   */
  public String[] getEphemeris () { return (ephemeris.clone()); }

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the number of frames in the file.
   *
   * @return the number of normal data frames, excluding any
   * pseudoframes.
   */
  public int getFrameCount () { return (framePosition.length); }

  ////////////////////////////////////////////////////////////

  /**
   * Gets the frame position of the specified frame.
   * 
   * @param frame the frame in the range [0..N-1].
   *
   * @return the file position in 16-bit words of the specified frame.
   * Pseudoframes are not counted as "normal" frames, so the position
   * will skip over any pseudoframes found.
   *
   * @see #getFrameCount
   */
  public int getFramePosition (
    int frame
  ) {

    return (framePosition[frame]);

  } // getFramePosition

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the data start time according to the QTrack start
   * pseudoframe.
   *
   * @return the data start time in milliseconds since the Unix epoch
   * (Jan 1, 1970, 00:00:00 UTC).
   */
  public long getStartTime () { return (startTime); }

  ////////////////////////////////////////////////////////////

  /** Tests this class. */
  public static void main (String[] argv) throws Exception {

    // Create file
    // -----------
    String name = argv[0];
    QuorumFile file = new QuorumFile (name, 11090);
    int frames = file.getFrameCount();
    System.out.println ("Opened Quorum file with " + frames + " frames");

    // Print start time
    // ----------------
    SimpleDateFormat format = new SimpleDateFormat ("yyyy/MM/dd HH:mm:ss z");
    format.setTimeZone (TimeZone.getTimeZone ("GMT"));
    System.out.println ("Start time = " + format.format (file.getStartTime()));
    
    // Print ephemeris
    // ---------------
    String[] elements = file.getEphemeris();
    System.out.println ("Got orbital elements =");
    for (int i = 0; i < 2; i++) System.out.println (elements[i]);

    // Print frame positions
    // ---------------------
    for (int i = 0; i < 10; i++)
      System.out.println ("Frame position " + i + " = " + 
        file.getFramePosition (i)*2);
    System.out.println ("...");
    for (int i = frames-1-10; i < frames-1; i++)
      System.out.println ("Frame position " + i + " = " + 
        file.getFramePosition (i)*2);

  } // main

  ////////////////////////////////////////////////////////////

} // QuorumFile class

////////////////////////////////////////////////////////////////////////
