////////////////////////////////////////////////////////////////////////
/*
     FILE: TelemetryStreamFactory.java
  PURPOSE: Create telemetry stream objects.
   AUTHOR: Peter Hollemans
     DATE: 2007/01/28
  CHANGES: 2010/12/18, PFH, added <?> to stream class variables to remove warnings
           2011/02/03, PFH, modified to use sequential stream class list

  Copyright (C) 2007-2011, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.telemetry;

// Imports
// -------
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.text.*;
import terrenus.orbit.*;
import terrenus.instrument.*;

/**
 * The <code>TelemetryStreamFactory</code> class creates {@link
 * TelemetryStream} objects from various different stream class
 * implementations.
 */
public class TelemetryStreamFactory {

  // Constants
  // ---------

  /** The resource file for telemetry stream class names. */
  private static final String PROPS_FILE = "streamClasses.properties";

  // Variables
  // ---------

  /** The list of telemetry stream class names. */
  private static List<String> streamList;

  /** The verbose mode flag. */ 
  private static boolean verboseMode;
  
  ////////////////////////////////////////////////////////////

  static {

    // Get streams from properties
    // ---------------------------
    InputStream stream = TelemetryStreamFactory.class.getResourceAsStream (
      PROPS_FILE);
    Properties props = new Properties();
    try { 
      props.load (stream); 
      SortedMap<String,String> streamMap = new TreeMap<String,String>();
      for (Object key : props.keySet())
        streamMap.put ((String) key, props.getProperty ((String) key));      
      streamList = new ArrayList<String> (streamMap.values());
    } // try
    catch (IOException e) { }
    finally { 
      try { stream.close(); }
      catch (IOException e) { }
    } // finally

  } // static

  ////////////////////////////////////////////////////////////

  /**
   * Sets the verbose mode flag.  In verbose mode, the results of
   * automatic file identification is printed.  Otherwise, automatic
   * file identification is performed silently.
   *
   * @param verboseMode the new verbose mode to use, TRUE to print
   * or FALSE to not print.
   */
  public static void setVerboseMode (boolean verboseMode) {
    
    TelemetryStreamFactory.verboseMode = verboseMode;
    
  } // setVerboseMode
  
  ////////////////////////////////////////////////////////////

  /** 
   * Creates a telemetry stream object from a file.
   *
   * @param fileName the data file name.  The file will be opened
   * using various different stream classes in the list until the
   * correct constructor is found.
   * 
   * @return a new stream object.
   *
   * @throws IOException if the stream could not be created.  Either
   * the file was not found, or the format was not recognized by any
   * stream.
   */
  public static TelemetryStream create (
    String fileName
  ) throws IOException {

    // Check file exists and is readable
    // ---------------------------------
    File file = new File (fileName);
    if (!file.canRead())
      throw new FileNotFoundException ("Cannot open " + fileName);
    
    // Loop through each stream class
    // ------------------------------
    Class<?>[] types = new Class[] {String.class};
    Object[] args = new Object[] {fileName};
    TelemetryStream stream = null;
    for (String streamName : streamList) {
      try {
        Class<?> streamClass = Class.forName (streamName);
        Constructor<?> constructor = streamClass.getConstructor (types);
        stream = (TelemetryStream) constructor.newInstance (args);
      } // try
      catch (Exception e) { 
        if (verboseMode) e.printStackTrace(); 
      } // catch
      if (stream != null) break;
    } // for

    // Give up and throw an error 
    // --------------------------
    if (stream == null) {
      throw new UnsupportedEncodingException ("Unknown telemetry stream in " + 
        fileName);      
    } // if

    return (stream);

  } // create

  ////////////////////////////////////////////////////////////

  /** Tests this class. */
  public static void main (String[] argv) throws Exception {

    // Create stream
    // -------------
    String name = argv[0];
    TelemetryStreamFactory.setVerboseMode (true);
    TelemetryStream stream = TelemetryStreamFactory.create (name);
    System.out.println ("Auto detected " + stream.getFormat() + 
      " telemetry stream");
    
    // Set stream year
    // ---------------
    String yearString = System.getProperty ("year");
    if (yearString != null)
      stream.setYear (Integer.parseInt (yearString));

    // Print frame summary
    // -------------------
    TelemetryDecoder decoder = stream.getDecoder();
    int frameCount = stream.getFrameCount();
    System.out.println ("Found " + frameCount + " frames"); 
    SimpleDateFormat format = new SimpleDateFormat ("yyyy/MM/dd HH:mm:ss z");
    format.setTimeZone (TimeZone.getTimeZone ("GMT"));
    System.out.println ("Start time = " + 
      format.format (new Date (decoder.getTime (0))));
    System.out.println ("End time = " + 
      format.format (new Date (decoder.getTime (frameCount-1))));

    // Print frame errors
    // ------------------
    List<Integer> errorList = new ArrayList<Integer>();
    for (int i = 0; i < frameCount; i++)
      if (decoder.hasErrors (i)) errorList.add (i);
    if (errorList.size() == 0)
      System.out.println ("Detected no frame errors");
    else {
      System.out.println ("Detected frame errors");
      int count = 1;
      for (int frame : errorList) {
        System.out.print (frame);
        if (count == 10) { System.out.println(); count = 1; }
        else { System.out.print (" "); count++; }
      } // for
      if (count != 1) System.out.println();
    } // else

    // Print spacecraft info
    // ---------------------
    Spacecraft craft = decoder.getSpacecraft();
    System.out.println ("Satellite: " + craft.getName());
    System.out.println ("Launch date: " + 
      format.format (craft.getLaunchDate()));

    // Print instrument info
    // ---------------------
    List<Instrument> instrumentList = craft.getInstruments();
    System.out.println ("Found " + instrumentList.size() + " instrument(s)");
    int count = 1;
    for (Instrument instr : instrumentList) {
      System.out.println ("Instrument " + count + ":");
      System.out.println ("  Long name = " + instr.getLongName());

      // Print radiometer info
      // ---------------------
      if (instr instanceof Radiometer) {
        Radiometer meter = (Radiometer) instr;
        int channels = meter.getChannelCount();
        System.out.println ("  Radiometer channels = " + channels);
        System.out.println ("  Samples per line = " + meter.getSampleCount());
        for (int i = 0; i < channels; i++) {
          System.out.println ("  Channel " + (i+1) + " = " + 
            meter.getChannelName (i+1) + (meter.isThermal (i+1) ? 
            " (thermal)" : " (visible)"));
        } // for
      } // if

    } // for

  } // main

  ////////////////////////////////////////////////////////////

} // TelemetryStreamFactory class

////////////////////////////////////////////////////////////////////////
