////////////////////////////////////////////////////////////////////////
/*
     FILE: hrptingest.java
  PURPOSE: Converts HRPT/CHRPT to other formats.
   AUTHOR: Peter Hollemans
     DATE: 2005/10/24
  CHANGES: 2007/01/28, PFH, modified to use automatic format detection
           2007/02/07, PFH, moved to tools package and added CHRPT support
           2007/03/14, PFH, completely reorganized class structure
           2009/06/01, PFH, added output format option
           2010/12/20, PFH, added raw CHRPT ingest, -d and -c options
           2011/01/13, PFH, added TeraScan CHRPT format

  Copyright (C) 2005-2011, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.tools;

// Imports
// -------
import java.util.*;
import java.io.*;
import java.text.*;
import jargs.gnu.*;
import jargs.gnu.CmdLineParser.*;
import terrenus.orbit.*;
import terrenus.telemetry.*;
import terrenus.instrument.*;
import terrenus.instrument.RadiometerCalibrator.CalibrationType;

/**
 * <p>The <code>hrptingest</code> program converts the
 * high-resolution radiometer sensor data in a NOAA HRPT or CMA CHRPT
 * raw capture station data file into calibrated, earth-located data
 * in HDF or NetCDF.</p>
 *
 * <h2>Synopsis</h2>
 *
 * <p>hrptingest [OPTIONS] input output</p>
 *
 * <h3>Options:</h3>
 *
 * <p>
 * -h, --help <br>
 * -i, --informat=TYPE <br>
 * -f, --outformat=TYPE <br>
 * -m, --minvalid=COUNT <br>
 * -o, --origin=STRING <br>
 * -V, --version <br>
 * -y, --year=YYYY <br>
 * -d, --day=DDD <br>
 * -c, --creationdate <br>
 * </p>
 *
 * <h2>Description</h2>
 *
 * <p>The <code>hrptingest</code> program converts the
 * high-resolution radiometer sensor data in a NOAA HRPT or CMA
 * CHRPT raw capture station data file into calibrated,
 * earth-located data in HDF or NetCDF.  Currently, only NOAA KLM
 * and CMA FY1 satellites are supported, including NOAA-15, -16,
 * -17, -18, -19, and FY-1D.  AVHRR or MVISR sensor data is
 * quality checked and re-calibrated every 5 lines.  In some
 * cases the output data may contain missing scan lines where the
 * corresponding telemetry frame data was corrupt.  The output
 * HDF or NetCDF file contains all AVHRR or MVISR channels
 * calibrated to scientific units, as well as earth locations and
 * ancilliary angles.  See the various {@link TelemetryStream}
 * classes for descriptions of the input formats, and the {@link
 * CWHDFDataWriter} and {@link CWNCDataWriter} classes for the
 * output format.</p>
 *
 * <h2>Parameters</h2>
 *
 * <h3>Main parameters:</h3>
 *
 * <dl>
 *
 *   <dt> input </dt>
 *   <dd> The input data file name. </dd>
 *
 *   <dt> output </dt>
 *   <dd> The output data file name. </dd>
 *
 * </dl>
 *
 * <h3>Options:</h3>
 *
 * <dl>
 *
 *   <dt>-h, --help</dt>
 *
 *   <dd>Prints a brief help message.</dd>
 *
 *   <dt>-i, --informat=TYPE</dt>
 *
 *   <Dd>The input format.  The supported formats are:
 *   <ul>
 *     <li>auto - Automatic format detection (default)</li>
 *     <li>quorum-hrpt - Quorum QTrack HRPT</li>
 *     <li>quorum-chrpt - Quorum QTrack CHRPT</li>
 *     <li>global - Global Imaging HRPT</li>
 *     <li>seaspace-hrpt - SeaSpace TeraScan HRPT</li>
 *     <li>seaspace-hrpt-swap - SeaSpace TeraScan HRPT byte-swapped</li>
 *     <li>seaspace-chrpt - SeaSpace TeraScan CHRPT</li>
 *     <li>raw-hrpt - Raw NOAA HRPT</li>
 *     <li>raw-chrpt - Raw CMA CHRPT</li>
 *     <li>1a5-hrpt - NOAA 1A5 HRPT</li>
 *     <li>1a5-chrpt - CMA 1A5 CHRPT</li>
 *   </ul></dd>
 *
 *   <dt>-f, --outformat=TYPE</dt>
 *
 *   <dd>The output format.  The supported formats are:
 *   <ul>
 *     <li>auto - Automatic format detection from file extension
 *     (default)</li>
 *     <li>hdf - Hierarchical Data Format (HDF)</li>
 *     <li>nc - Network Common Data Form (NetCDF)</li>
 *   </ul>  
 *
 *   <dt>-m, --minvalid=COUNT</dt>
 *
 *   <dd>Specifies the minimum number of consecutive valid data lines
 *   at the start and end of the data.  Data beyond this boundary is
 *   discarded.  The default is to write all data lines to the output
 *   file.  Invalid data lines are detected by examining the frame
 *   synchronization words and TIP data parity.</dd>
 *
 *   <dt>-o, --origin=STRING</dt>
 *
 *   <dd>Specifies the data origin string for the HDF 'origin'
 *   attribute.  The default is to write 'unknown' for the origin,
 *   which is not very informative.</dd>
 *
 *   <dt>-V, --version</dt>
 *
 *   <dd>Prints the software version number.</dd>
 *
 *   <dt>-y, --year=YYYY</dt>
 *
 *   <dd>Specifies the four digit year of data capture.  By default,
 *   the current year is assumed, unless the year can be determined 
 *   from the input data contents.  Currently, only the Quorum
 *   and 1A5 formats contain the year.</dd>
 *
 *   <dt>-d, --day=DDD</dt>
 *
 *   <dd>Specifies the three digit day of data capture (January 1 = 001).  
 *   By default, the current day is assumed, unless the day can be determined 
 *   from the input data contents.  Currently, only raw CHRPT requires
 *   a day to be specified.</dd>
 *
 *   <dt>-c, --creationdate</dt>
 *
 *   <dd>Specifies that the values for year and day of data capture (see -y and 
 *   -d above) should be obtained from the file creation date.  By default the
 *   current year and day are assumed unless the date can be determined from
 *   the input data contents.</dd>
 *
 * </dl>
 *
 * <h2>Exit status</h2>
 * <p>0 on success, > 0 on failure.  Possible causes of errors:
 * <ul>
 *   <li> Invalid command line option. </li>
 *   <li> Invalid input or output file names. </li>
 *   <li> Unrecognized format of input data. </li>
 *   <li> Unknown satellite code in input file. </li>
 *   <li> No applicable orbital elements available. </li>
 * </ul></p>
 *
 * <h2>Examples</h2>
 * <p>The following shows the ingest of HRPT data from a Quorum data file:
 * <pre>
 *   phollema@bean&lt;Data&gt; hrptingest -V
 *   [hrpt 2.0] hrptingest
 *   Java 1.5.0_07 on Mac OS X 10.4.9 i386
 *   phollema@bean&lt;Data&gt; hrptingest 2006-07-14-1905-n18.QuorumHRPT 2006-07-14-1905-n18.hdf       
 *   hrptingest: Opened Quorum QTrack HRPT telemetry data file with 4426 frames
 *   hrptingest: Data capture year is 2006
 *   hrptingest: Detected spacecraft noaa-18 launched 2005/05/20 JD 140
 *   hrptingest: Data starts at 2006/07/14 JD 195 19:05:25.245 UTC
 *   hrptingest: Data ends at   2006/07/14 JD 195 19:17:42.745 UTC
 *   hrptingest: Decoding data from frames 0 through 4425
 *   hrptingest: Writing avhrr radiometer channel and angle data
 * </pre>
 * The CoastWatch Utilities can then be used to query and analyze
 * the data:
 * <pre>
 *   phollema@bean&lt;Data&gt; cwinfo 2006-07-14-1905-n18.hdf
 *   Contents of file 2006-07-14-1905-n18.hdf
 * 
 *   Global information:
 *     Satellite:           noaa-18
 *     Sensor:              avhrr
 *     Date:                2006/07/14 JD 195
 *     Time:                19:05:25 UTC
 *     Scene time:          day
 *     Projection type:     swath
 *     Origin:              unknown
 *     Format:              CoastWatch HDF version 3.4
 * 
 *   Variable information:
 *     Variable       Type    Dimensions  Units          Scale     Offset   
 *     avhrr_ch1      short   4426x2048   percent        0.01      0        
 *     avhrr_ch2      short   4426x2048   percent        0.01      0        
 *     avhrr_ch3      short   4426x2048   celsius        0.01      0        
 *     avhrr_ch4      short   4426x2048   celsius        0.01      0        
 *     avhrr_ch5      short   4426x2048   celsius        0.01      0        
 *     avhrr_ch3a     short   4426x2048   percent        0.01      0        
 *     latitude       float   4426x2048   degrees        -         -        
 *     longitude      float   4426x2048   degrees        -         -        
 *     sat_zenith     short   4426x2048   degrees        0.01      0        
 *     sun_zenith     short   4426x2048   degrees        0.01      0        
 *     rel_azimuth    short   4426x2048   degrees        0.01      0        
 * 
 *   phollema@bean&lt;Data&gt; cwstats --sample 0.01 2006-07-14-1905-n18.hdf
 *   Variable       Count     Valid     Min        Max        Mean       Stdev    
 *   avhrr_ch1      90815     90815     1.93       106.77     15.970976  16.357994 
 *   avhrr_ch2      90815     90815     0.95       123.13     21.458364  15.816578 
 *   avhrr_ch3      90815     90815     -23.78     64.31      27.48028   13.601457 
 *   avhrr_ch4      90815     90815     -80.25     51.45      12.980522  20.530457 
 *   avhrr_ch5      90815     90815     -79.74     46.62      10.031554  20.063936 
 *   avhrr_ch3a     90815     0         NaN        NaN        NaN        NaN       
 *   latitude       90815     90815     13.350684  59.875427  37.012815  12.258589 
 *   longitude      90815     90815     -119.264236 -68.644142 -88.972349 8.879522  
 *   sat_zenith     90815     90815     0.21       69.11      32.621807  19.494969 
 *   sun_zenith     90815     90815     5.53       46         24.720583  7.579942  
 *   rel_azimuth    90815     90815     0          179.99     84.804809  60.840523
 * </pre>
 * </p>
 */
public class hrptingest {

  /** Minimum required command line parameters. */
  private static final int NARGS = 2;

  /** Name of program. */
  private static final String PROG = "hrptingest";

  /** Version of this software package. */
  private static final String VERSION = "2.2";

  ////////////////////////////////////////////////////////////

  /** Performs the main function. */
  public static void main (String[] argv) throws Exception {

    // Parse command line
    // ------------------
    CmdLineParser cmd = new CmdLineParser();
    Option helpOpt = cmd.addBooleanOption ('h', "help");
    Option informatOpt = cmd.addStringOption ('i', "informat");
    Option outformatOpt = cmd.addStringOption ('f', "outformat");
    Option originOpt = cmd.addStringOption ('o', "origin");
    Option yearOpt = cmd.addIntegerOption ('y', "year");
    Option dayOpt = cmd.addIntegerOption ('d', "day");
    Option creationdateOpt = cmd.addBooleanOption ('c', "creationdate");
    Option versionOpt = cmd.addBooleanOption ('V', "version");
    Option minvalidOpt = cmd.addIntegerOption ('m', "minvalid");
    cmd.parse (argv);

    // Print help message
    // ------------------
    Boolean helpObj = (Boolean) cmd.getOptionValue (helpOpt);
    if (helpObj != null && helpObj.booleanValue()) {
      usage();
      return;
    } // if

    // Print version number
    // --------------------
    Boolean versionObj = (Boolean) cmd.getOptionValue (versionOpt);
    if (versionObj != null && versionObj.booleanValue()) {
      System.out.println ("[hrpt " + VERSION + "] " + PROG);
      System.out.println (
        "Java " + System.getProperty ("java.version") + " on " +
        System.getProperty ("os.name") + " " + 
        System.getProperty ("os.version") + " " + 
        System.getProperty ("os.arch")
      );
      return;
    } // if

    // Get remaining arguments
    // -----------------------
    String[] remain = cmd.getRemainingArgs();
    if (remain.length < NARGS) 
      throw new RuntimeException ("At least " + NARGS + " arguments required");
    String input = remain[0];
    String output = remain[1];

    // Get options
    // -----------
    String informat = (String) cmd.getOptionValue (informatOpt);
    if (informat == null) informat = "auto";
    String outformat = (String) cmd.getOptionValue (outformatOpt);
    if (outformat == null) outformat = "auto";
    String origin = (String) cmd.getOptionValue (originOpt);
    GregorianCalendar calendar = new GregorianCalendar (TimeZone.getTimeZone ("GMT"));
    Integer yearObj = (Integer) cmd.getOptionValue (yearOpt);
    int year = (yearObj == null ? calendar.get (Calendar.YEAR) : yearObj.intValue());
    Integer dayObj = (Integer) cmd.getOptionValue (dayOpt);
    int day = (dayObj == null ? calendar.get (Calendar.DAY_OF_YEAR) : dayObj.intValue());
    Boolean creationdateObj = (Boolean) cmd.getOptionValue (creationdateOpt);
    boolean useCreationDate = (creationdateObj != null && creationdateObj.booleanValue());
    Integer minvalidObj = (Integer) cmd.getOptionValue (minvalidOpt);
    int minvalid = (minvalidObj == null ? 0 : minvalidObj.intValue());

    // Check input file
    // ----------------
    File inputFile = new File (input);
    if (!inputFile.exists()) 
      throw new RuntimeException ("Input file " + input + " does not exist");
    if (!inputFile.canRead())
      throw new RuntimeException ("Cannot access input file " + input);

    // Get input file date
    // -------------------
    if (useCreationDate) {
      Date fileDate = new Date (inputFile.lastModified());
      calendar.setTime (fileDate);
      year = calendar.get (Calendar.YEAR);
      day = calendar.get (Calendar.DAY_OF_YEAR);
    } // if

    // Get output file format
    // ----------------------
    if (outformat.equals ("auto")) {
      String fileName = output.toLowerCase();
      if (fileName.endsWith (".hdf")) outformat = "hdf";
      else if (fileName.endsWith (".nc")) outformat = "nc";
      else throw new RuntimeException ("Cannot determine output format");
    } // if
    if (!outformat.equals ("hdf") && !outformat.equals ("nc"))
      throw new RuntimeException ("Invalid output format specification");
    
    // Open input file
    // ---------------
    TelemetryStream stream;
    if (informat.equals ("auto")) {
      try { stream = TelemetryStreamFactory.create (input); }
      catch (IOException e) { stream = null; }
      if (stream == null)
        throw new IOException ("Cannot determine input file format");
    } // if
    else if (informat.equals ("global"))
      stream = new GlobalHRPTStream (input);
    else if (informat.equals ("seaspace") || informat.equals ("seaspace-hrpt"))
      stream = new TeraScanHRPTStream (input);
    else if (informat.equals ("seaspace-swap") || informat.equals ("seaspace-hrpt-swap"))
      stream = new TeraScanSwapHRPTStream (input);
    else if (informat.equals ("seaspace-chrpt"))
      stream = new TeraScanCHRPTStream (input);
    else if (informat.equals ("quorum") || informat.equals ("quorum-hrpt"))
      stream = new QuorumHRPTStream (input);
    else if (informat.equals ("quorum-chrpt"))
      stream = new QuorumCHRPTStream (input);
    else if (informat.equals ("raw") || informat.equals ("raw-hrpt"))
      stream = new RawHRPTStream (input);
    else if (informat.equals ("raw-chrpt"))
      stream = new RawCHRPTStream (input);
    else if (informat.equals ("1a5") || informat.equals ("1a5-hrpt"))
      stream = new CMA1A5HRPTStream (input);
    else if (informat.equals ("1a5-chrpt"))
      stream = new CMA1A5CHRPTStream (input);
    else 
      throw new RuntimeException ("Invalid input format specification");
    int frames = stream.getFrameCount();
    System.out.println (PROG + ": Opened " + stream.getFormat() + 
      " telemetry data file with " + frames + " frames");

    // Set stream year and date
    // ------------------------
    if (stream.getYear() == -1) stream.setYear (year);
    if (stream.getDayCount() == -1) stream.setDayCount (day);

    // Print spacecraft data
    // ---------------------
    TelemetryDecoder decoder = stream.getDecoder();
    SimpleDateFormat dateFormat = new SimpleDateFormat ("yyyy/MM/dd 'JD' DDD");
    dateFormat.setTimeZone (TimeZone.getTimeZone ("GMT"));
    Spacecraft craft = decoder.getSpacecraft();
    System.out.println (PROG + ": Detected spacecraft " + craft.getName() 
      + " launched " + dateFormat.format (craft.getLaunchDate()));

    // Print frame time range
    // ----------------------
    dateFormat.applyPattern("yyyy/MM/dd 'JD' DDD HH:mm:ss.SSS 'UTC'");
    System.out.println (PROG + ": Data starts at " + 
      dateFormat.format (new Date (decoder.getTime (0))));
    System.out.println (PROG + ": Data ends at   " + 
      dateFormat.format (new Date (decoder.getTime (frames-1))));

    // Prepare the history string
    // --------------------------
    String history = "[hrpt " + VERSION + "] " + PROG + " ";
    for (int i = 0; i < argv.length; i++) {
      history += argv[i];
      if (i != argv.length-1) history += " ";
    } // for

    // Find good start/end frames
    // --------------------------
    int startFrame = 0;
    int endFrame = frames-1;
    if (minvalid != 0) {

      // Find start frame
      // ----------------
      int goodFrames = 0;
      int frame;
      for (frame = 0; frame < frames && goodFrames < minvalid; frame++) {
        if (decoder.hasErrors (frame)) goodFrames = 0;
        else goodFrames++;
      } // for
      if (frame == frames)
        throw new RuntimeException ("Cannot locate minimum specified" +
          " valid lines");
      startFrame = frame - minvalid;

      // Find end frame
      // --------------
      goodFrames = 0;
      for (frame = frames-1; frame >= 0 && goodFrames < minvalid; frame--) {
        if (decoder.hasErrors (frame)) goodFrames = 0;
        else goodFrames++;
      } // for
      if (frame == -1)
        throw new RuntimeException ("Cannot locate minimum specified" +
        " valid lines");
      endFrame = frame + minvalid;

    } // if

    // Print start/end frames
    // ----------------------
    System.out.println (PROG + ": Decoding data from frames " + startFrame +
      " through " + endFrame);

    // Write output file
    // -----------------
    AbstractRadiometerDataWriter writer;
    if (outformat.equals ("hdf"))
      writer = new CWHDFDataWriter (output);
    else
      writer = new CWNCDataWriter (output);

    // FIXME: Figure out why the NetCDF writer writes 0 for uncalibrated data
    // rather than -32768.
    
    if (origin != null) writer.setOrigin (origin);
    writer.setHistory (history);
    Radiometer radiometer = (Radiometer) craft.getInstruments().get(0);
    System.out.println (PROG + ": Writing " + radiometer.getName() +
      " radiometer channel and angle data");
    writer.writeData (decoder, radiometer, startFrame, endFrame, 
      CalibrationType.ALBEDO, CalibrationType.CELSIUS);

  } // main

  ////////////////////////////////////////////////////////////

  /** Prints a usage message. */
  private static void usage () {

    System.out.println (
"Usage: " + PROG + " [OPTIONS] input output\n" +
"Converts HRPT/CHRPT data to a calibrated, earth-located dataset.\n" +
"\n" +
"Main parameters:\n" +
"  input                      The input data file name.\n" +
"  output                     The output data file name.\n" +
"\n" +
"Options:\n" + 
"  -h, --help                 Show this help message.\n" +
"  -i, --informat=TYPE        Set the input HRPT/CHRPT format.\n" +
"  -f, --outformat=TYPE       Set the output format.\n" +
"  -m, --minvalid=COUNT       Set the minimum valid start/end line count.\n" +
"  -o, --origin=STRING        Set the HDF origin attribute.\n" +
"  -V, --version              Print the software version number.\n" +
"  -y, --year=YYYY            Set the data year.\n" +
"  -d, --day=DDD              Set the data day (January 1 = 001).\n" +
"  -c, --creationdate         Use the input file creation date for year\n" +
"                             and day."
    );

  } // usage

  ////////////////////////////////////////////////////////////

} // hrptingest

////////////////////////////////////////////////////////////////////////
