////////////////////////////////////////////////////////////////////////
/*
     FILE: FrequencyMap.java
  PURPOSE: Tracks frequently occurring values.
   AUTHOR: Peter Hollemans
     DATE: 2007/02/08
  CHANGES: n/a

  Copyright (C) 2007, Terrenus Earth Sciences

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package terrenus.util;

// Imports 
// -------
import java.util.*;

/** 
 * The <code>FrequencyMap</code> class is used to store a number of values
 * (possible more than one occurrance of a value) and retrieve the most
 * frequently occurring value.
 */
public class FrequencyMap<E> {

  // Variables
  // ---------

  /** The map of value to frequency count. */
  private Map<E,int[]> valueMap = new HashMap<E,int[]>();

  ////////////////////////////////////////////////////////////

  /** 
   * Puts another occurrance of the value into the frequency map. 
   * 
   * @param value the value to add.
   */
  public void put (
    E value
  ) {

    int[] count = valueMap.get (value);
    if (count == null) valueMap.put (value, new int[] {1});
    else count[0]++;

  } // put

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the current most frequently occurring value.
   *
   * @return the most frequently occurring value or null if none
   * exists because the map is empty.
   */
  public E getFreqValue () {

    // Initialize search
    // -----------------
    E value = null;
    int maxCount = 0;

    // Look for value with maximum count
    // ---------------------------------
    for (Map.Entry<E,int[]> entry : valueMap.entrySet()) {
      int count = (entry.getValue())[0];
      if (count > maxCount) {
        maxCount = count;
        value = entry.getKey();
      } // if
    } // for

    return (value);

  } // getFreqValue

  ////////////////////////////////////////////////////////////

} // FrequencyMap class

////////////////////////////////////////////////////////////////////////

